import asyncio
import base64
import hashlib
import io
import json
import mimetypes
import os
import os.path
import shutil
import sys
import time
import urllib.request
import uuid
from io import BytesIO
from urllib.parse import urlparse
from urllib.request import Request, urlopen

import aiofiles
import aiohttp
import cloudinary
import cloudinary.uploader
import cv2
import ffmpeg
import numpy as np
import pydash
import pytesseract
import requests
from adafri.v1.files import File
from bs4 import BeautifulSoup
from cv2.gapi import cv
from google.cloud import storage
from PIL import Image
from PIL.Image import Image as PILImage
from werkzeug.utils import secure_filename

# def get_driver_options():
#     options = webdriver.ChromeOptions()
#     options.add_argument("--headless")
#     options.add_argument("--disable-gpu")
#     options.add_argument("window-size=1024,768")
#     options.add_argument("--no-sandbox")
#     options.add_argument('--ignore-certificate-errors')
#     options.add_argument("--test-type")
#     return options;

def get_storage_client():
    storage_credentials_path = os.environ.get('CLOUD_STORAGE_CREDENTIALS_PATH', None)
    storage_client = storage.Client.from_service_account_json(storage_credentials_path)
    return storage_client
def get_html_content(url, props, wait=1):
    # print('url', url)
    # print('url', props)
    #browser = webdriver.Chrome(service=Service(ChromeDriverManager().install()), options=get_driver_options());
    # browser.get(url);
    # browser.implicitly_wait(wait);
    # source = browser.page_source;
    # browser.close();
    source = urllib.request.urlopen(url);
    soup = BeautifulSoup(source, "html.parser");
    results = {};
    props_results = {};
    prop_key = 'key';
    prop_value = 'value';
    attr_key = 'attr';
    for prop in props:
        attrs = {};
        attrs[prop[attr_key]] = prop[prop_key]
        # print('find for', prop)
        # print('find for attr', prop[attr_key])
        props_results[prop[prop_key]]  = soup.find_all(attrs=attrs);
    for prop in props:
        if prop[prop_key] in props_results:
            for p in props_results[prop[prop_key]]:
                pv = prop[prop_value]
                # print(f"prop {prop[prop_key]} {pv}  result {p}");
                if p.has_attr(pv):
                    results[prop[prop_key]] = p[pv] 
    return results;

async def get_content_length(url):
    async with aiohttp.ClientSession() as session:
        async with session.head(url) as request:
            return request.content_length


def parts_generator(size, start=0, part_size=10 * 1024 ** 2):
    while size - start > part_size:
        yield start, start + part_size
        start += part_size
    yield start, size


async def download(url, headers, save_path):
    print(save_path)
    async with aiohttp.ClientSession(headers=headers) as session:
        async with session.get(url) as request:
            file = await aiofiles.open(save_path, 'wb')
            buffer = await request.content.read();
            await file.write(buffer)

async def downloadFile(file_input, save_path):
    # print(save_path)
    file = await aiofiles.open(save_path, 'wb')
    await file.write(file_input.read())


async def process(url):
    filename = os.path.basename(urlparse(url).path)
    #tmp_dir = TemporaryDirectory(prefix=filename, dir=f"{os.path.abspath('.')}/uploads/");
    size = await get_content_length(url)
    tasks = []
    file_parts = []

    for number, sizes in enumerate(parts_generator(size)):
        part_file_name = os.path.join(os.getcwd(), 'uploads/'+secure_filename(f'{filename}.part{number}'));
        file_parts.append(part_file_name)
        tasks.append(download(url, {'Range': f'bytes={sizes[0]}-{sizes[1]}'}, part_file_name))
    
    await asyncio.gather(*tasks)
    with open(os.path.join(os.getcwd(), 'uploads/'+filename), 'wb') as wfd:
        for f in file_parts:
            with open(f, 'rb') as fd:
                shutil.copyfileobj(fd, wfd)
                os.remove(f)

async def processFile(file):
    filename = file.filename
    #tmp_dir = TemporaryDirectory(prefix=filename, dir=f"{os.path.abspath('.')}/uploads/");
    size = len(file.read())
    tasks = []
    file_parts = []

    for number, sizes in enumerate(parts_generator(size)):
        part_file_name = os.path.join(os.getcwd(), 'uploads/'+secure_filename(f'{filename}.part{number}'));
        file_parts.append(part_file_name)
        tasks.append(downloadFile(file, part_file_name))
    
    await asyncio.gather(*tasks)
    
    with open(filename, 'wb') as wfd:
        for f in file_parts:
            with open(f, 'rb') as fd:
                shutil.copyfileobj(fd, wfd)
                os.remove(f)

async def main(url):
    await asyncio.gather(*[process(url)])

async def main_file(file):
    await asyncio.gather(*[processFile(file)])

def get_prob_attribute(url):
    hash = get_video_hash(url);
    print('video hash', hash);
    filename = os.path.basename(urlparse(url).path);
    directory_path = f"{os.path.abspath('.')}/uploads"
    abs_path = f"{directory_path}/{filename}"
    pre, ext = os.path.splitext(f"{directory_path}/{filename}")
    abs_path_without_ext = f"{pre}"    
    mime_type = "";
    extension = "";
    mpeg = ['mpeg','mpegts','mpegvideo']
    prob = ffmpeg.probe(abs_path)
    if prob['format']['format_name']=='mov,mp4,m4a,3gp,3g2,mj2':
        mime_type="video/mp4";
        extension = "mp4";
    elif prob['format']['format_name'] in mpeg:
        mime_type="video/mpeg";
        extension = prob['format']['format_name'];
    prob['type']="VIDEO"
    #print('prob', prob)

    prob['content_type'] = mime_type;
    prob['abs_path'] = abs_path;
    prob['abs_path_without_ext'] = abs_path_without_ext;
    prob['extension'] = extension;
    prob['filename'] = filename
    prob['directory_path'] = directory_path

    video_stream = next((stream for stream in prob['streams'] if stream['codec_type'] == 'video'), None)
    audio_stream = next((stream for stream in prob['streams'] if stream['codec_type'] != 'video'), None)
    if video_stream is not None:
        video_stream_attributes = ['avg_frame_rate', 'width', 'height', 'coded_width', 'coded_height', 'bit_rate', 'display_aspect_ratio', 'r_frame_rate', 'nb_frames', 'r_frame_rate', 'sample_aspect_ratio']
        video_prop = pydash.pick(video_stream, video_stream_attributes);
        video_prop['video_biterate'] = video_prop['bit_rate']
        del video_prop['bit_rate']
        prob = {**prob, **video_prop};

    del prob['format']['filename']
    prob['_hash'] = hash;
    return prob; 

def resize(image):
    maxsize = (1920, 1080)
    image.thumbnail(maxsize)


def resize_image_keep_ratio(image_path: str, target_width: int = None, target_height: int = None, 
    output_path: str = None, quality: int = 95) -> dict:
    """
    Resize image while maintaining aspect ratio.
    
    Args:
        image_path: Path to input image
        target_width: Target width (if None, height will be used)
        target_height: Target height (if None, width will be used)
        output_path: Output path (if None, overwrites original)
        quality: JPEG quality (1-100)
    
    Returns:
        Dict with original and new dimensions, file sizes
    """
    try:
        # Open image
        with Image.open(image_path) as img:
            # Get original dimensions
            original_width, original_height = img.size
            original_size = os.path.getsize(image_path)
            
            # Calculate new dimensions maintaining aspect ratio
            if target_width and target_height:
                # Both dimensions specified - fit within bounds
                ratio = min(target_width / original_width, target_height / original_height)
                new_width = int(original_width * ratio)
                new_height = int(original_height * ratio)
            elif target_width:
                # Width specified - calculate height
                ratio = target_width / original_width
                new_width = target_width
                new_height = int(original_height * ratio)
            elif target_height:
                # Height specified - calculate width
                ratio = target_height / original_height
                new_width = int(original_width * ratio)
                new_height = target_height
            else:
                return {"error": "Must specify either width or height"}
            
            # Resize image
            resized_img = img.resize((new_width, new_height), Image.Resampling.LANCZOS)
            
            # Determine output path
            if not output_path:
                output_path = image_path
            
            # Save resized image
            if output_path.lower().endswith('.jpg') or output_path.lower().endswith('.jpeg'):
                resized_img.save(output_path, 'JPEG', quality=quality, optimize=True)
            elif output_path.lower().endswith('.png'):
                resized_img.save(output_path, 'PNG', optimize=True)
            else:
                resized_img.save(output_path)
            
            # Get new file size
            new_size = os.path.getsize(output_path)
            
            return {
                "success": True,
                "original_dimensions": (original_width, original_height),
                "new_dimensions": (new_width, new_height),
                "original_size_kb": original_size // 1024,
                "new_size_kb": new_size // 1024,
                "size_reduction_percent": round((1 - new_size / original_size) * 100, 2),
                "output_path": output_path
            }
            
    except Exception as e:
        return {"error": f"Resize failed: {str(e)}"}

def resize_image_to_max_dimensions(image_path: str, max_width: int, max_height: int, 
    output_path: str = None, quality: int = 95) -> dict:
    """
    Resize image to fit within maximum dimensions while keeping aspect ratio.
    Only resizes if image exceeds the maximum dimensions.
    """
    try:
        with Image.open(image_path) as img:
            original_width, original_height = img.size
            
            # Check if resizing is needed
            if original_width <= max_width and original_height <= max_height:
                return {
                    "success": True,
                    "resized": False,
                    "message": "Image already within bounds",
                    "dimensions": (original_width, original_height)
                }
            
            # Calculate new dimensions
            ratio = min(max_width / original_width, max_height / original_height)
            new_width = int(original_width * ratio)
            new_height = int(original_height * ratio)
            
            # Resize
            resized_img = img.resize((new_width, new_height), Image.Resampling.LANCZOS)
            
            # Save
            if not output_path:
                output_path = image_path
                
            if output_path.lower().endswith('.jpg') or output_path.lower().endswith('.jpeg'):
                resized_img.save(output_path, 'JPEG', quality=quality, optimize=True)
            else:
                resized_img.save(output_path)
            
            return {
                "success": True,
                "resized": True,
                "original_dimensions": (original_width, original_height),
                "new_dimensions": (new_width, new_height),
                "output_path": output_path
            }
            
    except Exception as e:
        return {"error": f"Resize failed: {str(e)}"}

def replace_filename(_url: str, new_filename: str, _extension: str | None = None) -> str:
    last = _url.split('/')[-1]
    print('base', last)
    extension = last.split('.')[-1]
    if extension is None:
        extension = _extension
    file_name = last.split('/')[-1].split('.')[0]
    return _url.replace(last, f"{new_filename}_{file_name}.{extension}")


def get_image_attribute(url, max_width=1200, max_height=1200):
    im = Image.open(url);
    prob = {};
    width, height = im.size;
    # if width > max_width or height > max_height:
    #     print('resizing image', url)
    #     resized_url = replace_filename(url, "resized")
    #     resized = resize_image_to_max_dimensions(
    #         url, 
    #         max_width=max_width, 
    #         max_height=max_height, 
    #         output_path=resized_url
    #     )
    #     if resized  is not None and 'new_dimensions' in resized:
    #         width, height = resized['new_dimensions']
    #         print('new im after resize', width, height)
    #         im = Image.open(resized_url);
    #         im.save(url)
    #         os.remove(resized_url)

    filename = os.path.basename(urlparse(url).path);
    directory_path = f"{os.path.abspath('.')}/uploads"
    abs_path = f"{directory_path}/{filename}"
    pre, ext = os.path.splitext(f"{directory_path}/{filename}")
    abs_path_without_ext = f"{pre}"  
    size = os.stat(abs_path).st_size
    format = im.format;
    mime_type = str(format).lower();
    extension = str(format).lower();
    prob['width'] = width;
    prob['height'] = height
    prob['extension'] = extension
    prob['mime_type'] = mime_type
    jpg = ['jpeg', 'jpg']
    png = ["png"]
    gif = ["gif"]
    prob['mode'] = im.mode
    hash = None;
    if 'gif' in mime_type:
        hash = get_image_hash_pillow(url);
    else:
        hash = get_image_hash(url);
    if str(mime_type).lower() in jpg:
        prob['content_type'] = 'image/jpeg';
    elif mime_type in png:
        prob['content_type'] = 'image/jpeg';
    elif mime_type in gif:
        prob['content_type'] = 'image/gif';    #print('prob', prob)
    prob['content_type'] = mime_type;
    prob['abs_path'] = abs_path;
    prob['abs_path_without_ext'] = abs_path_without_ext;
    prob['extension'] = extension;
    prob['filename'] = filename
    prob['directory_path'] = directory_path
    prob['size'] = size
    prob['_hash'] = hash;
    return prob; 

def resumable_upload(url, remove=True):
    start_code = time.monotonic();
    asyncio.new_event_loop().run_until_complete(main(url))
    print(f'{time.monotonic() - start_code} seconds!')
    prob = get_prob_attribute(url);
    if remove is True:
        os.remove(prob['abs_path']);
    return prob;

def resumable_file_upload(file, extensions, upload_path, remove=True):
    name = file.filename;
    print('filename', file.filename)
    if file.filename == 'blob':
        ext = ".png";
        if file.content_type == 'image/jpeg':
            ext = ".jpg"
        name = name + ext
    filename = secure_filename(name)
    print('secure filename', secure_filename);
    path = os.getcwd() + "/"+ os.path.join(upload_path, filename);
    print(path)
    if filename != '':
        file_ext = os.path.splitext(filename)[1]
        if file_ext not in extensions:
            return None
        file.save(path)
    save_path = path

    prob = get_prob_attribute(path);
    if remove is True:
        os.remove(save_path);
    return prob;

def resumable_image_upload(file, extensions, fname=None, upload_path=None, remove=True):
    # name = None
    name = file.filename;
    if name == 'blob':
        name = f"file~{uuid.uuid1()}"
        ext = ".png";
        if file.content_type == 'image/jpeg':
            ext = ".jpg"
        name = name + ext

    if fname is not None and bool(fname):
        name = fname  

    filename = secure_filename(name);
    path = None
    if os.path.isabs(upload_path):
        path = os.path.join(upload_path, filename);
    else:
        path = os.path.join(os.getcwd(), upload_path, filename);
    # print('path', path)
    if filename != '':
        file_ext = getattr(file, 'extension', None)
        if file_ext is None:
            file_ext = os.path.splitext(filename)[1]
        # print('ext attr', ext)
        if file_ext not in extensions:
            print('file ext not in extensions', extensions)
            return None
        custom = getattr(file, 'is_custom', False)
        print('custom', custom)
        if getattr(file, 'is_custom', False) is True:
            file.image.save(path)
        else:
            file.save(path)
    save_path = path
    print('file saved on', save_path)

    prob = get_image_attribute(path);
    
    if remove is True:
        os.remove(save_path);
    return prob;


def generate_thumbnail(in_filename, out_filename):
    probe = ffmpeg.probe(in_filename)
    time = float(probe['streams'][0]['duration']) // 2
    width = probe['streams'][0]['width']
    try:
        (
            ffmpeg
            .input(in_filename, ss=time)
            .filter('scale', width, -1)
            .output(out_filename, vframes=1)
            .overwrite_output()
            .run(capture_stdout=True, capture_stderr=True)
        )
    except ffmpeg.Error as e:
        print(e.stderr.decode(), file=sys.stderr)
        sys.exit(1)


def crop_video(url, crop, save_key=uuid.uuid1()):
    ratio = 1
    print("crop", crop)
    if 'crop_based_width' in crop and crop['crop_based_width'] is not None:
        capture = cv2.VideoCapture(url)
        video_width = int(capture.get(cv.CAP_PROP_FRAME_WIDTH));
        ratio = crop['crop_based_width']/video_width;
        capture.release();
        
    x = int(crop['x']/ratio);
    y = int(crop['y']/ratio); 
    xPosition = int(crop['xPosition']/ratio);
    yPosition = int(crop['yPosition']/ratio);
    crop_format = {"x": x, "y": y, "xPosition": xPosition, "yPosition": yPosition};
    crop_hash = f"{hashlib.sha1(str(str(json.dumps(crop_format)) + '~' + save_key).encode()).hexdigest()}"
    prob = resumable_upload(url, False);
    #print(prob)
    outputName = f"{prob['abs_path_without_ext']}~{crop_hash}.{prob['extension']}";
    thumbnailName = f"{prob['abs_path_without_ext']}~{crop_hash}.jpg";
    #print('output', outputName)
    crop = ffmpeg.input(prob['abs_path']).filter('crop', xPosition, yPosition, x, y).output(outputName).run();
    crop_prob = get_prob_attribute(outputName);
    thumbnail = generate_thumbnail(outputName, thumbnailName)
    print('crop prob', crop_prob)
    crop_prob['parent_path'] = prob['abs_path']
    crop_prob['thumbnail'] = thumbnailName
    #os.remove(prob['abs_path'])
    return crop_prob;

def crop_image_fn(url, crop):
    crop_hash = hashlib.sha1(json.dumps(crop).encode()).hexdigest();
    im = Image.open(requests.get(url, stream=True).raw);
    format = im.format;
    extension = str(format).lower();
    mime_type = str(format).lower();
    content_type = ""
    jpg = ['jpeg', 'jpg']
    png = ["png"]
    gif = ["gif"]
    if str(mime_type).lower() in jpg:
        content_type = 'image/jpeg';
    elif mime_type in png:
        content_type = 'image/png';
    elif mime_type in gif:
        content_type = 'image/gif';
    filename_hash = hashlib.sha1(str(f"crop~{os.path.basename(urlparse(url).path)}~{crop_hash}").encode()).hexdigest();
    filename = f"{filename_hash}.{extension}" 
    directory_path = f"{os.path.abspath('.')}/uploads"
    abs_path = f"{directory_path}/{filename}"
    pre, ext = os.path.splitext(f"{directory_path}/{filename}")
    abs_path_without_ext = f"{pre}"
    image_width, image_height = im.size;
    if str(ext).lower() == 'png':
        extension = "PNG"  
    crop_based_width = image_width;
    if 'crop_based_width' in crop and crop['crop_based_width'] is not None:
        crop_based_width = crop['crop_based_width']
    ratio = crop_based_width/image_width;
    # print('ratio',  ratio)
    x = int(crop['x']/ratio);
    y = int(crop['y']/ratio); 
    xPosition = int(crop['xPosition']/ratio);
    yPosition = int(crop['yPosition']/ratio);

    im.crop((x, y, xPosition, yPosition)).save(abs_path, extension, optimize=True);
    image_cropped = Image.open(abs_path)
    st_size = os.stat(abs_path).st_size
    prob = {
        'width': image_cropped.width,
        'height': image_cropped.height,
        'extension': extension,
        'size': st_size,
        'mime_type': mime_type,
        'type': "IMAGE",
        'content_type': content_type,
        'abs_path': f"{abs_path_without_ext}.{extension}",
        'abs_path_without_ext': abs_path_without_ext,
        'filename': filename
    }
    return prob;


def hash_array_to_hash_hex(hash_array):
  # convert hash array of 0 or 1 to hash string in hex
  hash_array = np.array(hash_array, dtype = np.uint8)
  hash_str = ''.join(str(i) for i in 1 * hash_array.flatten())
  hash_hex = (hex(int(hash_str, 2)));
  hash = hashlib.sha1(str(hash_hex).encode()).hexdigest();
  return hash

def hash_hex_to_hash_array(hash_hex):
  # convert hash string in hex to hash values of 0 or 1
  hash_str = int(hash_hex, 16)
  array_str = bin(hash_str)[2:]
  arr = np.array([i for i in array_str], dtype = np.float32)
  string = np.array2string(arr);
  hash = hashlib.sha1(str(hash).encode()).hexdigest();
  print('hash', hash);
  return hash;


def get_image_hash(image):
    img = cv2.imread(image);
  # resize image and convert to gray scale
    img = cv2.resize(img, (64, 64))
    img = cv2.cvtColor(img, cv2.COLOR_BGR2GRAY)
    img = np.array(img, dtype = np.float32)
    # calculate dct of image 
    dct = cv2.dct(img)
    # to reduce hash length take only 8*8 top-left block 
    # as this block has more information than the rest
    dct_block = dct[: 8, : 8]
    # caclulate mean of dct block excluding first term i.e, dct(0, 0)
    dct_average = (dct_block.mean() * dct_block.size - dct_block[0, 0]) / (dct_block.size - 1)
    # convert dct block to binary values based on dct_average
    dct_block[dct_block < dct_average] = 0.0
    dct_block[dct_block != 0] = 1.0
    # store hash value
    hash = hash_array_to_hash_hex(dct_block.flatten());
    print('hash', hash);
    return hash;
def get_image_hash_pillow(image):
    img = Image.open(image)
    img = img.resize((10, 10), Image.ANTIALIAS)
    img = img.convert("L")
    pixel_data = list(img.getdata())
    avg_pixel = sum(pixel_data)/len(pixel_data)
    bits = "".join(['1' if (px >= avg_pixel) else '0' for px in pixel_data])
    hex_representation = str(hex(int(bits, 2)))[2:][::-1].upper();
    hash = hashlib.sha1(str(hex_representation).encode()).hexdigest();
    print('hash', hash);
    return hash;
def get_video_hash(video):
    print(video)
    cap = cv2.VideoCapture(video)
    frames = [];
    hashes = [];
    check = True
    i = 0
    bgst = cv2.createBackgroundSubtractorMOG2()


    while check:
        check, frame = cap.read()
        # if not i % 20: 
    # When everything done, release the capture
        # np_arr = np.array(frames);
        #img = cv2.resize(frame, (64, 64))
        if frame is not None:
            img = cv2.cvtColor(frame, cv2.COLOR_BGR2GRAY)
            img = np.array(img, dtype = np.float32)
            dct = cv2.dct(img)
            # to reduce hash length take only 8*8 top-left block 
            # as this block has more information than the rest
            dct_block = dct[: 8, : 8]
            # caclulate mean of dct block excluding first term i.e, dct(0, 0)
            dct_average = (dct_block.mean() * dct_block.size - dct_block[0, 0]) / (dct_block.size - 1)
            # convert dct block to binary values based on dct_average
            dct_block[dct_block < dct_average] = 0.0
            dct_block[dct_block != 0] = 1.0
            # store hash value
            hash = hash_array_to_hash_hex(dct_block.flatten());
            #print('frame hash', hash)
            hashes.append(hash)
        # print('check', check) # This line is if you want to subsample your video
        # print('arr', arr)         # (i.e. keep one frame every 20)
        # frames.append(arr)
        #i += 1
    cap.release()

    hash_data = hashlib.sha1(str(''.join(hashes)).encode()).hexdigest();
    #print('hash', hash_data);
    return hash_data;


def text_detection(input):
    image = None
    try:
        if isinstance(input, str) is True:
            image = Image.open(input);
        elif input is not None:
            image = input;
        text = pytesseract.image_to_string(image);
        #print the text line by line
        print(text[:-1])
        print(text)
        return text[:-1].splitlines();
    except Exception as e:
        return None;

def analyze(urls=[]):
    try:
        url = f"{ANALYZE_API_URL}/analyzeImage"
        data_object = {'images_url': urls}
        x = requests.post(url, json = data_object)
        data = x.json()
        if len(data)==0:
            return None;
        return data[0]['colors'];
    except Exception as e:
        return None;

def get_content_type(_format: str):
    format = _format.lower()
    if format == 'jpeg':
        format = 'jpg'
    return f"image/{format}"

def upload_blob_from_blob_From_File(file_source_name, bucket_name,  destination_blob_name, project_id):
    #os.environ["GOOGLE_APPLICATION_CREDENTIALS"] = "adafri-backend-google-b205ae995e46.json"
    storage_client = get_storage_client()
    bucket = storage_client.bucket(bucket_name)
    blob = bucket.blob(destination_blob_name)
    # local_size_kb = os.stat(file_source_name).st_size // 1024
    #print(bucket.cors)
    image = Image.open(file_source_name)
    # mime_type = image.get_format_mimetype()
    # print('mime_type', mime_type)
    # st_size = os.stat(file_source_name).st_size
    # print('stat size', st_size)
    cropped_bytes = image.tobytes()
    # local_size_kb = len(cropped_bytes) // 1024
    # blob.upload_from_string(
    #         cropped_bytes,
    #         content_type=mime_type,
    #         timeout=300
    # )
    use_cloudinary = as_bool(os.environ.get('USE_CLOUDINARY', False))
    if use_cloudinary:
        up = cloudinary_upload(file_source_name, bucket_name, destination_blob_name)
        if up is None:
            return None
        return up
    blob.upload_from_filename(file_source_name, timeout=300)
    
    blob.make_public();
    # blob.reload()
    gcs_size_kb = blob.size
    # print('local_size_kb', local_size_kb)
    # print('blob.size', gcs_size_kb // 1024)
    return {"file_url": blob.public_url, "bucket_name": bucket_name, "size": gcs_size_kb, "blob_name": blob.name, "created_time": blob.time_created.strftime("%Y%m%d"), "blob_id": blob.id, "content_type": blob.content_type}


def cloudinary_upload(filename, bucket_name='adafri_upload', blob_name='image_upload'):
    try:
        print('cloudinary upload', filename)
        up = cloudinary.uploader.upload(filename)
        if up is None:
            return None
        up['file_url'] = up['secure_url']
        up['bucket_name'] = bucket_name
        up['blob_name'] = blob_name
        up['content_type'] = get_content_type(up['format'])
        up['size'] = up['bytes']
        print('cloudinary upload response', up)
        return up
    except Exception as e:
        print('cludinary exception upload', e)
        return None
def as_bool(val):
    if isinstance(val, bool):
        return val
    return val.lower() in ('true', '1')


class FileWrapper(PILImage):
    image: PILImage
    is_custom = True
    extension = ''
    content_type = ''
    def __init__(self, file_path=None, filename=None, image_data=None):
        super().__init__()
        self.file_path = file_path
        self.filename = filename
        self.image_data = image_data
        self.is_custom = True
        if image_data is not None:
            self.file_size = len(image_data)
    
    def get_format(self, value: str):
        if value.lower() == 'jpeg':
            return '.jpg'
        return f'.{value}'
    def open(self):
        """Open and validate the image data"""
        try:
            # Create image from bytes without verification (which can be destructive)
            image = Image.open(io.BytesIO(self.image_data))
            
            # Get image format and set extension
            im_format = image.format
            if im_format:
                file_extension = self.get_format(im_format)
                self.extension = file_extension
                print('file_extension', file_extension)
                
                # Set content type
                if im_format.upper() == 'PNG':
                    self.content_type = 'image/png'
                elif im_format.upper() in ['JPEG', 'JPG']:
                    self.content_type = 'image/jpeg'
                elif im_format.upper() == 'GIF':
                    self.content_type = 'image/gif'
                else:
                    self.content_type = f'image/{im_format.lower()}'
            else:
                # Default to jpg if format can't be determined
                self.extension = '.jpg'
                self.content_type = 'image/jpeg'
            
            self.image = image
            return self
            
        except Exception as e:
            print('Exception while openeing', e)
            return None
            # raise ValueError(f"Failed to open image: {e}")
    
    def save(self, path):
        """Save the image data to a file"""
        try:
            with open(path, 'wb') as f:
                f.write(self.image_data)
        except Exception as e:
            raise ValueError(f"Failed to save image: {e}")
    
    # def read(self):
    #     with open(self.file_path, 'rb') as f:
    #         return f.read()


def upload_image_from_string(image_string, extensions, upload_path, accountId: str, public=True, process=True, name=None, bucket_name='adafri_upload', upload_folder='image_upload', project_id="adafri-backend") -> 'File | None':
    """
    Upload image from a data URL string (e.g., "data:image/png;base64,iVBORw0KGgo...")
    
    Args:
        image_string: Data URL string starting with "data:image/..."
        extensions: Allowed file extensions
        upload_path: Path to save the image
        accountId: User account ID
        public: Whether to make the file public
        process: Whether to process the image
        name: Custom filename
        bucket_name: Cloud storage bucket name
        upload_folder: Folder within the bucket
        project_id: Google Cloud project ID
    
    Returns:
        File object or None if upload fails
    """
    try:
        # Parse the data URL
        if not image_string.startswith('data:image/'):
            raise ValueError("Invalid data URL format. Must start with 'data:image/'")
        
        # Extract MIME type and base64 data
        header, encoded_data = image_string.split(',', 1)
        mime_type = header.split(':')[1].split(';')[0]
        
        # Get file extension from MIME type
        extension = mimetypes.guess_extension(mime_type)
        if not extension:
            extension = '.png'  # Default to PNG if can't determine
        
        # Remove the dot from extension
        extension = extension.lstrip('.')
        
        # Validate extension
        if f".{extension.lower()}" not in [ext.lower() for ext in extensions]:
            raise ValueError(f"File extension '{extension}' not allowed. Allowed: {extensions}")
        
        # Decode base64 data
        try:
            image_data = base64.b64decode(encoded_data)
        except Exception as e:
            raise ValueError(f"Invalid base64 data: {str(e)}")
        
        # Validate image data
        
        # Generate filename if not provided
        if not name:
            timestamp = int(time.time())
            name = f"image_{timestamp}_{hashlib.md5(image_data).hexdigest()[:8]}"
        
        # Create temporary file path
        temp_filename = f"{name}.{extension}"
        name = temp_filename
        temp_file_path = os.path.join(upload_path, temp_filename)
        image = None
        try:
            file_wrapper = FileWrapper(temp_file_path, temp_filename, image_data)
            image = file_wrapper.open()
            width, height = image.size
        except Exception as e:
            raise ValueError(f"Invalid image data: {str(e)}")
        
        # Ensure upload directory exists
        os.makedirs(upload_path, exist_ok=True)
        
        # Save image to temporary file
        with open(temp_file_path, 'wb') as f:
            f.write(image_data)
        
        # Create a file-like object for the upload function
        
        # Use existing upload function
        result = upload_image_from_file(
            file=image,
            extensions=extensions,
            upload_path=upload_path,
            accountId=accountId,
            public=public,
            process=process,
            name=name,
            bucket_name=bucket_name,
            upload_folder=upload_folder,
            project_id=project_id
        )
        
        # Clean up temporary file
        try:
            os.remove(temp_file_path)
        except Exception as e:
            print(f"Warning: Could not remove temporary file {temp_file_path}: {e}")
        
        return result
        
    except Exception as e:
        print(f"Error uploading image from string: {str(e)}")
        return None
def upload_image_from_file(file, extensions, upload_path, accountId: str, public=True, process=True, name=None, bucket_name='adafri_upload', upload_folder='image_upload', project_id="adafri-backend") ->'File | None':
    # Use the bucket to get a blob object;
    #upload = upload_file(file, accountId, False, True, bucket_name)
    resumable = resumable_image_upload(file=file, extensions=extensions, fname=name, upload_path=upload_path, remove=False);
    if resumable is None:
        return None
    
    filenameHash = f"{hashlib.sha1(str(resumable['abs_path_without_ext'] + '~' + accountId).encode()).hexdigest()}";
    filename = f"{filenameHash}.{resumable['extension']}"
    blob_name = f"{upload_folder}/{accountId}/{filename}";
    obj_str = blob_name + '~'+ str(resumable['_hash']) + '~' + accountId;
    use_cloudinary = as_bool(os.environ.get('USE_CLOUDINARY', False))
    if 'size' in resumable:
        obj_str = obj_str + '~' + str(resumable['size'])
    if 'width' in resumable:
        print('width resumable', resumable['width'])
        obj_str = obj_str + '~' + str(resumable['width'])
    if 'height' in resumable:
        obj_str = obj_str + '~' + str(resumable['height'])
    if use_cloudinary:
        obj_str = obj_str + '~' + os.getenv('CLOUDINARY_API_KEY','cloudinary')


    # obj_id = f"{hashlib.sha1(str(str(resumable['_hash']) + '~' + accountId).encode()).hexdigest()}";
    # print('resumable', resumable)
    obj_id = f"{hashlib.sha1(str(obj_str).encode()).hexdigest()}"
    file = File(file=obj_id, collection_name=f"{upload_folder}").get();
    if file is not None:
        return file
    print('uploading start')
    size = 0
    url = ""
    blob_bucket_name = bucket_name
    blob_id=None
    type=None
    created_time=None
    collection = upload_folder
    file_path = f"{resumable['directory_path']}/{resumable['filename']}"
    
    print('use_cloudinary', use_cloudinary)
    if use_cloudinary:
        print('upload_image_from_file with cloudinary', file_path)
        up = cloudinary_upload(file_path, bucket_name, blob_name)
        if up is None:
            print("cloudinary exception upload", up)
            return None
        up['id'] = obj_id
        up['_id'] = obj_id
        up['obj_id'] = obj_id
        type = up['content_type']
        size = up['size']
        url = up['file_url']
    else:
        print('upload_image_from_file with gcloud', file_path)
        storage_client = get_storage_client()
        bucket = storage_client.bucket(bucket_name)
        blob = bucket.blob(blob_name)
        try:
            os_file_path = file_path
            #os.path.join(os.getcwd(), 'uploads/'+resumable['filename']);
            blob.upload_from_filename(filename=os_file_path)
        except Exception as e:
            print("Exception catched", e);
            return {"error": e}
        # Make the object public
        if public:
            blob.make_public()
        url = blob.public_url
        size = blob.size
        blob_bucket_name = blob.bucket.name
        type = blob.content_type
        blob_name = blob.name
        blob_id = blob.id
        created_time = blob.time_created
   
    analyzed = None;
    text_detected = None;
    if process is True:
        try:
            analyzed = analyze([url]);
            text_detected = text_detection(file_path);
        except Exception as e:
            print("Exception catched", e);
    try:
        os.remove(file_path);
    except Exception as e:
            print("Exception catched", e);
    #blob_split  = blob.id.split('/')
    #blob_id = blob_split[len(blob_split)-1]; 
    video_obj = {"file_source": None, "file_url": url, "bucket_name": blob_bucket_name, "accountId": accountId, "size": size, "blob_name": blob_name, "created_time": created_time, "blob_id": blob_id, "type": type, **resumable}
    video_info = pydash.omit(video_obj, ['abs_path', 'abs_path_without_ext', 'parent_path'])
    if analyzed is not None and len(analyzed)>0:
        video_info['colors'] = analyzed;
    if text_detected is not None and len(text_detected)>0:
        video_info['text'] = text_detected;
    #obj_id = f"{hashlib.sha1(str(blob.name + '~' + str(blob.size) + '~' + accountId).encode()).hexdigest()}"
    video_info['_id'] = obj_id
    video_info['id'] = obj_id
    # video_info['testMode'] = True 
    # print('video obj', video_info)
    file_create = File(collection_name=f"{upload_folder}").create(**video_info);
    # print('file_create', file_create)
    if file_create is None or file_create.status.lower()!="ok":
        return None
    # db = FirebaseCollectionBase(collection_name=collection).db
    # db.collection(collection).document(obj_id).set(video_info, merge=True)
    # return video_info
    return file_create.data
def upload_image_from_url(file, extensions, upload_path, accountId: str, public=True, process=True, name=None, bucket_name='adafri_upload', upload_folder='image_upload', project_id="adafri-backend") ->'File':
    # Use the bucket to get a blob object;
    #upload = upload_file(file, accountId, False, True, bucket_name)
    image = None;
    try:
        # os.makedirs(f"{upload_path}/{accountId}", exist_ok=True)
        image = get_image_bytes(file)
    except Exception as e:
        print("Exception catched", e);
    collection = upload_folder
    # print('image', image)
    if image is None:
        return None
        
    storage_client = get_storage_client()
    bucket = storage_client.bucket(bucket_name)
    image_bytes, pic = image
    im_format = pic.format
    width, height = pic.size
    if name is None:
        name = f"image-{uuid.uuid1()}"
    file_extension = '';
    if f".{im_format.lower()}" in  extensions:
        file_extension = "." + im_format.lower()
    filename = f"{name}{file_extension}"
    blob_name = f"{upload_folder}/{accountId}/{filename}";
    blob = bucket.blob(blob_name)
    # print("Image format is: ", im_format)
    content_type = 'image/jpeg'
    public = ''
    if im_format=='PNG':
        content_type = 'image/png'
        #pic.convert('RGB')
    obj_str = str(file) + '~' + str(content_type) + '~' + str(blob.size) + '~' + str(width) + '~' + str(height) + '~' + accountId;
    obj_id = f"{hashlib.sha1(str(obj_str).encode()).hexdigest()}"
    file = File(file=obj_id, collection_name=f"{upload_folder}").get();
    if file is not None:
        return file
    blob.upload_from_string(image_bytes, content_type=content_type)
    blob.make_public()
    public = blob.public_url
    # print(blob.metadata)

    if public:
        blob.make_public()
   
    analyzed = None;
    text_detected = None;
    if process is True:
        try:
            analyzed = analyze([blob.public_url]);
            text_detected = text_detection(pic);
        except Exception as e:
            print("Exception catched", e);

    #blob_split  = blob.id.split('/')
    #blob_id = blob_split[len(blob_split)-1]; 
    video_obj = {"file_source": file, "file_url": blob.public_url, "bucket_name": blob.bucket.name, "accountId": accountId, "size": blob.size, "blob_name": blob.name, "created_time": blob.time_created, "blob_id": blob.id, "type": blob.content_type, "width": width, "height": height}
    video_info = pydash.omit(video_obj, ['abs_path', 'abs_path_without_ext', 'parent_path'])
    if analyzed is not None and len(analyzed)>0:
        video_info['colors'] = analyzed;
    if text_detected is not None and len(text_detected)>0:
        video_info['text'] = text_detected;
    #obj_id = f"{hashlib.sha1(str(blob.name + '~' + str(blob.size) + '~' + accountId).encode()).hexdigest()}"
    video_info['_id'] = obj_id
    video_info['id'] = obj_id
    file_create = File(file=video_info, collection_name=f"{upload_folder}").create(**video_info);
    if file_create is None or file_create.status.lower()!="ok":
        return None
    # db = FirebaseCollectionBase(collection_name=collection).db
    # db.collection(collection).document(obj_id).set(video_info, merge=True)
    # return video_info
    return file_create.data


def crop_image(url: str, accountId: str, crop, upload_folder="image_upload", bucket_name="adafri_upload", project_id="adafri-backend"):
    crop_image_data = crop_image_fn(url, crop);
    # print('crop_image_data', crop_image_data)
    file_path = f"{upload_folder}/{accountId}/{crop_image_data['filename']}"
    upload = upload_blob_from_blob_From_File(crop_image_data['abs_path'], bucket_name, file_path, project_id)
    os.remove(crop_image_data['abs_path']);
    print('crop_image_data', crop_image_data)
    print('upload', upload)
    result = pydash.omit({**crop_image_data, **upload}, ['abs_path', 'abs_path_without_ext', 'parent_path']);
    return result

def crop_image_from_url(url: str, accountId: str, crop, upload_folder="image_upload", bucket_name="adafri_upload", id=None, project_id="adafri-backend")->'File | None':
    data = {
        "bucket_name": bucket_name,
        "accountId": accountId,
        "upload_folder": upload_folder,
        "url": url,
        "file_url": url,
        "crop": crop,
    }
    
    file = None
    if id is not None:
        file = File(file=id, collection_name=f"{upload_folder}").get();
        if file is not None:
            if file.url is not None and file.file_url is None:
                file.file_url = file.url
            if file.crop_state == "done":
                return file
    value = crop_image(url, accountId, crop, upload_folder, bucket_name, project_id)
    result = pydash.omit({**value, **data}, ['abs_path', 'abs_path_without_ext', 'parent_path']);
    # video_hash = hashlib.sha1(json.dumps(result).encode()).hexdigest();
    crop_state = "done";
    #result['id'] = video_hash;
    result['crop_state'] = crop_state
    file_create = File(file=result, collection_name=f"{upload_folder}").create(**result);
    if file_create is None or file_create.status.lower()!="ok":
        return None
    # db = FirebaseCollectionBase(collection_name=collection).db
    # db.collection(collection).document(obj_id).set(video_info, merge=True)
    # return video_info
    return file_create.data
    # db.collection(upload_folder).document(doc_init['id']).set({**result}, merge=True)
    # return doc_init | result;
def get_image_bytes(url: str):
    headers = {
        "User-Agent": "Mozilla/5.0"
    }
    resp = requests.get(url, headers=headers, timeout=10)
    resp.raise_for_status()  # Raises an error for bad status codes

    # Check if the response is actually an image
    content_type = resp.headers.get("Content-Type", "")
    if not content_type.startswith("image/"):
        raise ValueError(f"URL did not return an image. Content-Type: {content_type}")

    # Try to open the image
    try:
        # print('rsp', resp.content)
        image_bytes = resp.content
        img = Image.open(BytesIO(resp.content))
        img.verify()  # Verify that it is, in fact, an image
        # If you want to return the image object, reopen it (verify() closes the file)
        img = Image.open(BytesIO(resp.content))
        return image_bytes,img 
    except Exception as e:
        raise ValueError(f"Failed to identify image file: {e}")
    # image_bytes = urlopen(url).read()
    # print("Image size is: ", image_bytes)
    # try:
    #     pic = Image.open(BytesIO(image_bytes))
    #     im_format = pic.format
    #     print("Image format is: ", im_format)
    #     # content_type = 'image/jpeg'
    #     # public = ''
    #     # if im_format=='PNG':
    #     #     content_type = 'image/png'
    #     return image_bytes, pic
    # except Exception as e:
    #     print("Exception catched", e);
    #     return None

def get_compression_estimate(size_kb: int, target_size_kb: int) -> float:
    """
    Get compression ratio needed to achieve target file size.
    Returns compression ratio (0.0 to 1.0, where 1.0 = no compression).
    """
    if not size_kb or size_kb <= 0:
        return 1.0
    
    compression_ratio = target_size_kb / size_kb
    return max(0.1, min(1.0, compression_ratio))

def estimate_pre_crop_file_size(width: int, height: int, current_size_kb: int, target_width: int, target_height: int) -> int:
    """
    Estimate the file size in KB needed before cropping to achieve target dimensions.
    
    Args:
        width, height: Current image dimensions
        current_size_kb: Current file size in KB
        target_width, target_height: Desired crop dimensions
    
    Returns:
        Estimated file size in KB needed before cropping
    """
    try:
        if not width or not height or not current_size_kb:
            return current_size_kb or 0
        
        # Calculate current and target areas
        current_area = width * height
        target_area = target_width * target_height
        
        if current_area <= 0 or target_area <= 0:
            print('current_area <= 0 or target_area <= 0')
            return current_size_kb
        
        # Calculate the ratio of target area to current area
        area_ratio = target_area / current_area
        
        # If target area is larger than current area, we need a bigger image
        if area_ratio > 1:
            # Estimate size needed for the larger area
            estimated_size = int(current_size_kb * area_ratio)
        else:
            print('area_ratio <= 1')
            # If target is smaller, current size should be sufficient
            estimated_size = current_size_kb
        
        return estimated_size
        
    except Exception as e:
        print(f"Error estimating pre-crop file size: {e}")
        return current_size_kb or 0

def calculate_required_image_size(target_width: int, target_height: int, current_width: int, current_height: int) -> tuple[int, int]:
    """
    Calculate the minimum image size needed to achieve target crop dimensions.
    
    Returns:
        Tuple of (required_width, required_height)
    """
    try:
        if not target_width or not target_height:
            return (current_width, current_height)
        
        # Calculate aspect ratios
        target_ratio = target_width / target_height
        current_ratio = current_width / current_height
        
        if target_ratio > current_ratio:
            # Target is wider - need to increase width
            required_width = int(current_height * target_ratio)
            required_height = current_height
        else:
            # Target is taller - need to increase height
            required_width = current_width
            required_height = int(current_width / target_ratio)
        
        return (required_width, required_height)
        
    except Exception as e:
        print(f"Error calculating required image size: {e}")
        return (current_width, current_height)

def estimate_cropped_file_size(width: int, height: int, current_size_kb: int, crop_x: int, crop_y: int, crop_width: int, crop_height: int) -> int:
    """
    Estimate the file size of the cropped portion before actually cropping.
    
    Args:
        width, height: Original image dimensions
        current_size_kb: Original file size in KB
        crop_x, crop_y: Crop starting coordinates
        crop_width, crop_height: Crop dimensions
    
    Returns:
        Estimated file size in KB of the cropped portion
    """
    try:
        if not width or not height or not current_size_kb:
            return 0
        
        # Calculate original and crop areas
        original_area = width * height
        crop_area = crop_width * crop_height
        
        if original_area <= 0 or crop_area <= 0:
            return 0
        
        # Calculate the ratio of crop area to original area
        area_ratio = crop_area / original_area
        
        # Estimate file size based on area ratio
        # File size is roughly proportional to image area
        estimated_crop_size = int(current_size_kb * area_ratio)
        
        return estimated_crop_size
        
    except Exception as e:
        print(f"Error estimating cropped file size: {e}")
        return 0

def estimate_crop_size_from_coordinates(image_width: int, image_height: int, image_size_kb: int, crop_data: dict) -> int:
    """
    Estimate cropped file size from crop coordinates.
    
    Args:
        image_width, image_height: Original image dimensions
        image_size_kb: Original file size in KB
        crop_data: Dict with x, y, width, height coordinates
    
    Returns:
        Estimated file size in KB of the cropped portion
    """
    try:
        crop_x = crop_data.get('x', 0)
        crop_y = crop_data.get('y', 0)
        crop_width = crop_data.get('width', 0)
        crop_height = crop_data.get('height', 0)
        
        return estimate_cropped_file_size(
            image_width, image_height, 
            image_size_kb, 
            crop_x, crop_y, crop_width, crop_height
        )
        
    except Exception as e:
        print(f"Error estimating crop size from coordinates: {e}")
        return 0

# # Usage example:
# original_width, original_height = 1200, 800
# original_size_kb = 800
# crop_coords = {'x': 100, 'y': 100, 'width': 600, 'height': 400}

# estimated_crop_size = estimate_crop_size_from_coordinates(
#     original_width, original_height,
#     original_size_kb, crop_coords
# )

# print(f"Original: {original_width}x{original_height}, {original_size_kb}KB")
# print(f"Crop area: {crop_coords['width']}x{crop_coords['height']}")
# print(f"Estimated cropped file size: {estimated_crop_size}KB")
# # Usage example:
# current_width, current_height = 800, 600
# current_size_kb = 500
# target_width, target_height = 400, 300

# # Estimate size needed before cropping
# estimated_size = estimate_pre_crop_file_size(
#     current_width, current_height, 
#     current_size_kb, 
#     target_width, target_height
# )

# # Calculate required dimensions
# required_width, required_height = calculate_required_image_size(
#     target_width, target_height,
#     current_width, current_height
# )

# print(f"Current: {current_width}x{current_height}, {current_size_kb}KB")
# print(f"Target crop: {target_width}x{target_height}")
# print(f"Required image: {required_width}x{required_height}")
# print(f"Estimated size needed: {estimated_size}KB")


def get_file_size_from_url(url: str) -> int:
    """
    Get accurate file size in KB from URL by downloading in chunks.
    Returns size in KB, or 0 if unable to determine.
    """
    try:
        # Stream the response to get accurate size
        with requests.get(url, stream=True, timeout=30) as response:
            response.raise_for_status()
            
            # Get content length if available
            content_length = response.headers.get('content-length')
            if content_length:
                size_kb = int(content_length)
                return size_kb
            
            # If no content-length, calculate by reading in chunks
            total_size = 0
            chunk_size = 8192  # 8KB chunks
            
            for chunk in response.iter_content(chunk_size=chunk_size):
                if chunk:
                    total_size += len(chunk)
                    # Stop after reading 1MB to avoid downloading huge files
                    if total_size > 1024 * 1024:  # 1MB limit
                        break
            
            return total_size  # Convert to KB
            
    except Exception as e:
        print(f"Error getting file size from URL: {e}")
        return 0

def get_file_size_accurate(url: str) -> int:
    """
    Alternative method using range requests for more accurate size.
    """
    try:
        # Try to get file size using range request
        headers = {'Range': 'bytes=0-1023'}  # Request first 1KB
        response = requests.get(url, headers=headers, timeout=10)
        
        if response.status_code == 206:  # Partial content
            # Parse Content-Range header
            content_range = response.headers.get('content-range', '')
            if '/' in content_range:
                total_size = int(content_range.split('/')[-1])
                return total_size
        
        # Fallback to streaming method
        return get_file_size_from_url(url)
        
    except Exception as e:
        print(f"Error getting accurate file size: {e}")
        return 0
def get_file_info_from_url(url: str) -> dict:
    """
    Get comprehensive file information from URL.
    Returns dict with size, content_type, and other headers.
    """
    try:
        response = requests.head(url, timeout=10, allow_redirects=True)
        response.raise_for_status()
        
        content_length = response.headers.get('content-length', '0')
        size_kb = int(content_length) if content_length != '0' else 0
        
        return {
            'size_kb': size_kb,
            'content_type': response.headers.get('content-type', ''),
            'content_length': content_length,
            'last_modified': response.headers.get('last-modified', ''),
            'etag': response.headers.get('etag', '')
        }
    except Exception as e:
        print(f"Error getting file info from URL: {e}")
        return {'size_kb': 0, 'content_type': '', 'content_length': '0'}


import random
import string
import time
from dataclasses import dataclass
from datetime import datetime
from typing import Any

import pydash
from adafri.cache.lib import CacheClass
from firebase_admin.firestore import firestore

from .....utils import (
    ArrayUtils,
    BaseClass,
    DateUtils,
    DictUtils,
    RequestFields,
    get_object_model_class,
    get_request_fields,
    init_class_kwargs,
)
from .....utils.response import ApiResponse, Error, ResponseStatus, StatusCode
from .....v1 import version
from ....account import Account
from ....base.firebase_collection import FirebaseCollectionBase
from ...data.demographics import ages_data, genders_data
from ...data.devices import devices_data
from ...models.base import BaseCampaign, BaseCampaignFieldsProps
from ...models.client_customer_id import ClientCustomerId
from ...models.placement import Website, YoutubeChannel, YoutubeVideo
from ...models.user_interest import UserInterest
from ..adgroups import AdGroupDisplay
from ..ads import ImageAd, ImageAdToPublish, NativeAdsToPublish
from ..ads import ResponsiveDisplayAd as NativeAd
from .display_fields import (
    CAMPAIGN_DISPLAY_NOT_PICKABLE_FIELDS,
    CAMPAIGN_DISPLAY_PICKABLE_FIELDS,
    DISPLAY_COLLECTION,
    CampaignDisplayFields,
    CampaignDisplayFieldsProps,
)


def filter_campaign_display_request_fields(fields, default_fields = CAMPAIGN_DISPLAY_PICKABLE_FIELDS):
    request_fields = get_request_fields(fields, default_fields, [default_fields[0]])
    if request_fields is None or bool(request_fields) is False:
        request_fields = [default_fields[0]]
    return request_fields



@dataclass(init=False)
class CampaignDisplay(BaseCampaign):
    id: str
    targetedPlacements: list[Website]
    excludedPlacements: list[Website]
    websites: list[Website]
    youtubeChannels: list[YoutubeChannel]
    youtubeVideos: list[YoutubeVideo]
    images: list[ImageAdToPublish]
    imagesNative: list[NativeAdsToPublish]
    user_interest: list[UserInterest]
    target_google_search: bool
    target_search_network: bool
    target_content_network: bool
    target_partner_search_network: bool
    __baseFields = CampaignDisplayFieldsProps
    class_object = None
    ad_group_class = None;

    def __init__(self, campaign=None, **kwargs):
        _campaign = campaign;
        if type(campaign) is str:
            _campaign = {'id': campaign};
        (cls_object, keys, data_args) = init_class_kwargs(self, _campaign, CAMPAIGN_DISPLAY_PICKABLE_FIELDS, CampaignDisplayFieldsProps, DISPLAY_COLLECTION, ['id'], **kwargs)
        super().__init__(**{**data_args, "campaign": _campaign});
        for key in keys:
            if BaseCampaignFieldsProps.get(key) is None:
                website = ['targetedPlacements', 'excludedPlacements', 'websites']
                if key in website:
                    setattr(self, key, Website().fromListDict(cls_object[key]))
                elif key == 'youtubeChannels':
                    setattr(self, key, YoutubeChannel().fromListDict(cls_object[key]))
                elif key == 'youtubeVideos':    
                    setattr(self, key, YoutubeVideo().fromListDict(cls_object[key]))
                elif key == 'images':
                    setattr(self, key, ImageAdToPublish().fromListDict(cls_object[key]))
                elif key == 'imagesNative':
                    setattr(self, key, NativeAdsToPublish().fromListDict(cls_object[key]))
                elif key == 'user_interest':
                    setattr(self, key, UserInterest().fromListDict(cls_object[key]))
                else:
                    setattr(self, key, cls_object[key]) 
        self.class_object = CampaignDisplay
        self.ad_group_class = AdGroupDisplay
        self.cache = CacheClass("campaign_display", CampaignDisplay)

    def get_local_responsive_display_ads(self):
        values = [];
        print('len', len(self.imagesNative))
        for image in self.imagesNative.copy():
            values.append(image.to_responsive_display_ad(self.owner))
        return values
    def get_local_image_ads(self):
        values = [];
        for image in self.images.copy():
            values.append(image.to_image_ad(self.owner))
        return values
    def build_adgroup(self, testMode = False):
        if self.is_published() is True or self.api_version!="v0":
            return None
        campaign = CampaignDisplay.from_dict(self.to_dict())
        adgroup = AdGroupDisplay().create({"name": f"Adgroup for {self.name}"}, campaign, testMode)
        if adgroup.status == ResponseStatus.OK:
            print("adgroup migrate success", f"{adgroup.data.id} => {adgroup.data.name}")
            return AdGroupDisplay.from_dict(adgroup.data.to_dict())
        elif adgroup.status == ResponseStatus.ERROR and adgroup.error.name == "DUPLICATE_ENTITY":
            print("adgroup migrate success already exist", f"{adgroup.data.id} => {adgroup.data.name}")
            return AdGroupDisplay.from_dict(adgroup.data.to_dict())
        else:
            print("adgroup migrate error", adgroup.error.message)
        return None
    def migrate_ads(self, testMode = False) -> 'tuple[AdGroupDisplay, list[ImageAd], list[NativeAd]]':
        adgroup = self.build_adgroup(testMode);
        if adgroup is None:
            return None, [], [];
        image_ads, native_ads = self.clone_local_ads(adgroup, testMode);
        return adgroup, image_ads, native_ads
    def build_local_native_ads(self, adgroup: AdGroupDisplay, testMode = False)-> 'list[NativeAd]':
        local_responsive: list[NativeAd] = self.get_local_responsive_display_ads();
        print('local natives', len(local_responsive))
        ads_to_build = [];
        for ad in local_responsive:
            generated = NativeAd().generateAd(adgroup, ad.to_dict(), testMode);
            if generated.status == ResponseStatus.OK:
                ad_data = generated.data;
                ad_data.ad_id = 0
                ads_to_build.append(ad_data)
            else:
                print('local responsive ad error',generated.error.message)
        return ads_to_build
    def build_local_image_ads(self, adgroup: AdGroupDisplay, testMode = False)-> 'list[ImageAd]':
        local_ads: list[ImageAd] = self.get_local_image_ads();
        ads_to_build = [];
        for ad in local_ads:
            generated = ImageAd().generateAd(adgroup, ad.to_dict(), testMode);
            if generated.status == ResponseStatus.OK:
                ad_data = generated.data;
                ad_data.ad_id = 0
                ads_to_build.append(ad_data)
            else:
                print('local image ad error',generated.error.message)

        return ads_to_build

    def clone_local_ads(self, adgroup: AdGroupDisplay, testMode = False) -> 'tuple[list[ImageAd], list[NativeAd]]':
        if self._ads_cloned is True:
            return [], []
        if self.is_published() is True:
            return [], []
        # image_ads = self.build_local_image_ads(adgroup, testMode);
        image_ads = None;
        native_ads = self.build_local_native_ads(adgroup, testMode);
        # for ad in image_ads:
        operations = None
        if testMode is True:
            return image_ads, native_ads
        if bool(image_ads) is True:
            batch_add_image = ImageAd().write_batch(image_ads);
            if batch_add_image.status == ResponseStatus.ERROR:
                operations = False
            # return batch_add.data
            else:
                print('write batch image ads success')
                for ad in batch_add_image.data:
                    print("new image ad created", f"{ad.id} => {ad.name} on adgroup {adgroup.id} and campaign {self.id}")
                operations = True
        if bool(native_ads) is True:
            batch_add_native = NativeAd().write_batch(native_ads);
            if batch_add_native.status == ResponseStatus.ERROR:
                operations = False
            else:
                print('write batch native ads success')
                for ad in batch_add_native.data:
                    print("new native ad created", f"{ad.id} => {ad.name} on adgroup {adgroup.id} and campaign {self.id}")
                operations = True
        
        if operations is True:
            self.update({"_ads_cloned": True})
        return image_ads, native_ads


    def get_adgroup_class():
        return AdGroupDisplay
    def placements_to_list(self, devices: list[Website] | None = None, fields = None):
        if devices is None:
            return []
        return Website().toListDict(devices, fields)
    def channels_to_list(self, channels: list[YoutubeChannel] | None = None, fields = None):
        if channels is None:
            return []
        return YoutubeChannel().toListDict(channels, fields)
    def videos_to_list(self, videos: list[YoutubeVideo] | None = None, fields = None):
        if videos is None:
            return []
        return YoutubeVideo().toListDict(videos, fields)
    def interest_to_list(self, interest: list[UserInterest] | None = None, fields = None):
        if interest is None:
            return []
        return UserInterest().toListDict(interest, fields)
    def list_to_placements(self, placements: list[Website] | None = None, fields = None):
        if placements is None:
            return []
        return Website().toListDict(placements, fields)
    def list_to_channels(self, channels: list[YoutubeChannel] | None = None, fields = None):
        if channels is None:
            return []
        return YoutubeChannel().toListDict(channels, fields)
    def list_to_videos(self, videos: list[YoutubeVideo] | None = None, fields = None):
        if videos is None:
            return []
        return YoutubeVideo().toListDict(videos, fields)
    def list_to_interests(self, interest: list[UserInterest] | None = None, fields = None):
        if interest is None:
            return []
        return UserInterest().toListDict(interest, fields)
    @staticmethod
    def generate_model(_key_="default_value"):
        model = {};
        props = CampaignDisplayFieldsProps
        for k in DictUtils.get_keys(props):
            if k=='ages':
                model[k] = ages_data;
                continue
            if k=='genders':
                model[k] = genders_data;
                continue
            if k=='devicesTargeted':
                model[k] = devices_data;
                continue
            if k=='api_version':
                model[k] = version
                continue
            if _key_ in props[k] and props[k][_key_] is not None:
                model[k] = props[k][_key_];
        
        # Generate random campaign name
        random_suffix = ''.join(random.choices(string.ascii_uppercase + string.digits, k=6))
        model['name'] = f'Campaign Display {random_suffix}'
        model['startDate'] = datetime.now().strftime('%Y-%m-%d')
        return model;

    @staticmethod
    def from_dict(campaign: Any, db=None, collection_name=None) -> 'CampaignDisplay':
        cls_object, keys = get_object_model_class(campaign, CampaignDisplay, CampaignDisplayFieldsProps);
        _campaign = CampaignDisplay(cls_object, db=db, collection_name=collection_name)
        return _campaign
    
    def get_customer_id(self, _customerId: str | None = None):
        customer = None;
        if _customerId is None:
            default_customers = ClientCustomerId().getDefault()
            if(len(default_customers) > 0):
                customer = default_customers[0]
        else:
            customer = ClientCustomerId(_customerId).get()
        if customer is None:
            return ApiResponse(ResponseStatus.ERROR, StatusCode.status_400, None, Error("Customer not found","INVALID_REQUEST", 1))
        return ApiResponse(ResponseStatus.OK, StatusCode.status_200, customer, None)
    def generateCampaign(self, aacid: str, _customerId: str | None = None, agroup=None, data: dict | None=None):
        account = Account(aacid).get()
        if account is None:
            return ApiResponse(ResponseStatus.ERROR, StatusCode.status_400, None, Error("Account not found","INVALID_REQUEST", 1))
        
        customer_request = self.get_customer_id(_customerId);
        if customer_request.status == 'error':
            return customer_request
        customer = customer_request.data
        # token = guards(['profile'])
        # if token is None:
        #     return ApiResponse(ResponseStatus.ERROR, StatusCode.status_400, None, Error("Account not found","PERMISSION_DENIED", 1)).to_json(), StatusCode.status_400
        model = CampaignDisplay.generate_model()
        model['accountId'] = aacid
        model['clientCustomerId'] = customer.customerId
        if data is not None:
            model = DictUtils.merge_dict(data, model, True)
        create = CampaignDisplay().create(model, account, account.owner, agroup)
        if create.status == 'error':
            return create
        # print('fields', CAMPAIGN_DISPLAY_PICKABLE_FIELDS)
        values = create.data
        values = DictUtils.omit_fields(values, CAMPAIGN_DISPLAY_NOT_PICKABLE_FIELDS + ['owner', 'id_campagne', 'ad_group_id'])
        # print('values', values)
        return ApiResponse(ResponseStatus.OK, StatusCode.status_200, values, None)
    def to_dict(self, fields=None):
        base = super().to_dict(fields)
        return DictUtils.remove_none_values({
            **base,
            "targetedPlacements": Website().toListDict(self.targetedPlacements, None),
            "excludedPlacements": Website().toListDict(self.excludedPlacements, None),
            "websites": Website().toListDict(self.websites, None),
            "youtubeChannels": YoutubeChannel().toListDict(self.youtubeChannels, None),
            "youtubeVideos": YoutubeVideo().toListDict(self.youtubeVideos, None),
            "images": ImageAdToPublish().toListDict(self.images, None),
            "imagesNative": NativeAdsToPublish().toListDict(self.imagesNative, None),
            "user_interest": UserInterest().toListDict(self.user_interest, None),
            "target_google_search": self.target_google_search,
            "target_search_network": self.target_search_network,
            "target_content_network": self.target_content_network,
            "target_partner_search_network": self.target_partner_search_network
    })
    def getCampaigns(self, aacid, fields=None):
        campaigns = []
        campaigns_ = self.query(CampaignDisplay, "campaign", query_params=[{"key": "accountId", "comp": "==", "value": aacid}])
        request_fields = filter_campaign_display_request_fields(fields)
        # print('account query', account)
        for campaign in campaigns_:
            campaigns.append(campaign)
            # campaigns.append(D(campaign.to_dict(request_fields)))

        return campaigns
    @staticmethod
    def findCampaign(id, aacid=None):
        isID = False
        try:
            id = int(id)
        except Exception as e:
            isID = True
        if isID is True:
            campaign = CampaignDisplay.from_dict({"id": id}).get(CAMPAIGN_DISPLAY_PICKABLE_FIELDS)
            if campaign is None:
                return None
            return campaign 
        if id is None or aacid is None or (isinstance(id, str) and len(id) == 0) or (isinstance(aacid, str) and len(aacid)) == 0 or (isinstance(id, int) and id==0):
            return None
        campaign = CampaignDisplay(None).query(CampaignDisplay, "campaign", query_params=[{"key": "accountId", "comp": "==", "value": aacid}, {"key": "id_campagne", "comp": "==", "value": id}], first=True)
        return campaign
    def handle_migrate(self):
        adgroup, image_ads, native_ads = self.migrate_ads(False);
        if bool(adgroup) is True:
            if bool(image_ads) is True:
                print("campaign have image ads migrate", f"{len(image_ads)} image ads")
            if bool(native_ads) is True:
                print("campaign have native ads migrate", f"{len(native_ads)} native ads")
    def get(self, fields=None):
        if bool(self.id) is None:
            return None;
        campaign_cache: CampaignDisplay | None = self.cache.get(self.id);
        if bool(campaign_cache) is True:
            return campaign_cache
        request_fields = filter_campaign_display_request_fields(fields)
        doc = self.document_reference(self.id).get();
        if doc.exists is False:
            return None;
        data = {"id": doc.id, **doc.to_dict()}
        campaign = CampaignDisplay.from_dict(campaign=data, db=self.db, collection_name=self.collection_name);
        if bool(campaign) is False:
            return None
        self.cache.set(campaign.id, campaign);
        # print('created at===>', DateUtils.convert_firestore_timestamp_to_str(data['createdAt']))
        return campaign;

    def getByUserId(self,id):
        campaigns = [];
        docs = self.collection().where('owner','==',id).get();
        for doc in docs:
            data = doc.to_dict();
            data['id'] = doc.id;
            campaigns.append(CampaignDisplay.from_dict(data))
        return campaigns;

    @staticmethod
    def fields_extra():
        return [
            CampaignDisplayFields.budgetEnded,
            CampaignDisplayFields.deliveryMethod,
            CampaignDisplayFields.is_removed,
            CampaignDisplayFields.isComplete,
            CampaignDisplayFields.publish,
            CampaignDisplayFields.publishing,
            CampaignDisplayFields.publicationDate,
            CampaignDisplayFields.isEnded,
            CampaignDisplayFields.isPayed,
            CampaignDisplayFields.servingStatus,
        ]
    def normalize_response(self, _fields=None):
        fields = ['owner', 'createdBy']
        if _fields is not None:
            fields.extend(_fields)
        data = CampaignDisplay.from_dict(DictUtils.pick_fields(self.to_dict(), CAMPAIGN_DISPLAY_PICKABLE_FIELDS))
        return CampaignDisplay.from_dict(DictUtils.omit_fields(data.to_dict(), ["owner", "createdBy"]))

    def create(self, data, account: Any, uid, _adgroup=None):
        # creation_date = created_at * 1000
        # return DateUtils.from_timestamp(creation_date).isoformat()
        # try:
        data_create = DictUtils.dict_from_keys(data, CampaignDisplayFields().filtered_keys('mutable', True));
        if bool(data_create) is False:
            return ApiResponse(ResponseStatus.ERROR, StatusCode.status_400, None, Error("Invalid request","INVALID_REQUEST", 1))
        campaign_dates = DateUtils.format_campaign_date(data_create)
        if campaign_dates.status == "error":
            return campaign_dates
        model = CampaignDisplay.generate_model()
        data_create = DictUtils.merge_dict({
            **campaign_dates.data,
                "accountId": account.id,
                "owner": account.owner,
                "createdBy": f"users/{uid}"
            }, data_create,True)
        data_create = DictUtils.merge_dict(data_create, model, True)
        document_id = self.collection().document().id

        exist = self.query(CampaignDisplay, "campaign", [{"key": "name", "comp": "==", "value": data_create['name']}], True)
        if exist is not None:
            return ApiResponse(ResponseStatus.ERROR, StatusCode.status_400, None, Error(f"Campaign name {data_create['name']} already exists","INVALID_REQUEST", 1))
        model = CampaignDisplay.generate_model()
        customerId = None
        if 'clientCustomerId' in data_create:
            customerId = data_create['clientCustomerId']
        customer_request = self.get_customer_id(customerId);
        if customer_request.status == 'error':
            return customer_request
        customer = customer_request.data
        data_create['clientCustomerId'] = customer.customerId
        model['id'] = document_id
        data_create = DictUtils.merge_dict(data_create, model, True)
        create = {**data_create, 'createdAt': firestore.SERVER_TIMESTAMP}
        doc_ref = self.collection().document(document_id);
        doc_ref.set(create, True);
        data_campaign = CampaignDisplay.from_dict(create)
        campaign = data_campaign.normalize_response(CampaignDisplay.fields_extra())
        adgroup = None;
        if bool(_adgroup) is True:
            adgroup_values = {"name": "Adgroup " + campaign.name, **AdGroupDisplay.get_ad_group_props(data_create)}
            if type(_adgroup) is dict:
                adgroup_values = DictUtils.merge_dict(_adgroup, adgroup_values, True)
            adgroup = AdGroupDisplay().generateAdgroup(data_campaign, adgroup_values)
            if adgroup.status == ResponseStatus.ERROR:
                return adgroup.to_json(), adgroup.status_code
        if campaign is None:
            return ApiResponse(ResponseStatus.ERROR, StatusCode.status_400, None, Error("Failed to create campaign","INVALID_REQUEST", 1))
        # print('adgroup', adgroup.data)
        response_values = {"campaign": campaign};
        if bool(adgroup) is True:
            response_values['adgroup'] = adgroup.data
        return ApiResponse(ResponseStatus.OK, StatusCode.status_200, response_values, None);  # Return the created campaign object.  # Return the created campaign object.  # Return the created campaign object.  # Return the created campaign object.  # Return the created campaign object.  # Return the created campaign object.  # Return the created campaign object.  # Return the created campaign object.  # Return the created campaign object.  # Return the created campaign object.  # Return the created campaign object.  # Return the
        # except Exception as e:
        #     print(e)
        #     return {"error": "Exception creating campaign"};
    
    def dates_keys(self):
        return [
            "startDate",
            "endDate",
            "startDateFormattedGoogle",
            "endDateFormattedGoogle"
        ]
    def prepareAddCriterion(self, data):
        campaign_add: CampaignDisplay = self.keep_only(CampaignDisplay.from_dict(data.copy()), pydash.keys(data.copy()));
        data_add = {};
        data_final = {};
        current_targeted = self.locations_to_list(self.targetedLocations).copy();
        current_excluded = self.locations_to_list(self.excludedLocations).copy();
        current_schedules = self.schedules_to_list(self.adsSchedules).copy();
        print('current schedules', current_schedules)
        targeted_locations = [];
        excluded_locations = [];
        schedules = [];
        valid_campaign_status = ["enabled", "paused"]

        if bool(campaign_add.name) is True and bool(campaign_add.name.strip()) is True:
            data_add['name'] = campaign_add.name;
            data_final['name'] = campaign_add.name;
        if bool(campaign_add.status) is True and campaign_add.status.lower() in valid_campaign_status:
            data_add['status'] = campaign_add.status;
            data_final['status'] = campaign_add.status;
        if bool(campaign_add.startDate) or bool(campaign_add.endDate):
            print('dates', campaign_add.startDate, campaign_add.endDate)
            dates = DateUtils.format_campaign_date(data);
            if dates.status == "ok":
                dates_data = pydash.pick(dates.data, self.dates_keys());
                print('dates_data', dates.data)
                data_add = DictUtils.merge_dict(dates_data, data_add, True);
                data_final = DictUtils.merge_dict(dates_data, data_final, True);

        if bool(campaign_add.targetedLocations) is True:
            values_to_add = self.locations_to_list(campaign_add.targetedLocations);
            if bool(current_targeted) is True:
                for value in values_to_add:
                    index = pydash.find_index(current_excluded, lambda x: x['criterionId']==value['criterionId']);
                    if index == -1:
                        if pydash.find(current_targeted, lambda x: x['criterionId']==value['criterionId']) is None:
                            if pydash.find(targeted_locations, lambda x: x['criterionId']==value['criterionId']) is None:
                                targeted_locations.append(value);
            else:
                for value in values_to_add:
                    index = pydash.find_index(current_excluded, lambda x: x['criterionId']==value['criterionId']);
                    if index == -1:
                        targeted_locations.append(value);
        
        
        if bool(campaign_add.excludedLocations) is True:
            values_to_add = self.locations_to_list(campaign_add.excludedLocations);
            if bool(current_excluded) is True:
                for value in values_to_add:
                    index = pydash.find_index(current_targeted, lambda x: x['criterionId']==value['criterionId']);
                    if index == -1:
                        if pydash.find(current_excluded, lambda x: x['criterionId']==value['criterionId']) is None:
                            if pydash.find(excluded_locations, lambda x: x['criterionId']==value['criterionId']) is None:
                                excluded_locations.append(value);
            else:
                for value in values_to_add:
                    index = pydash.find_index(current_targeted, lambda x: x['criterionId']==value['criterionId']);
                    if index == -1:
                        excluded_locations.append(value);

        if bool(campaign_add.adsSchedules) is True:
            values_to_add = self.schedules_to_list(campaign_add.adsSchedules);
            if bool(current_schedules) is True:
                print('current sch', current_schedules)
                print('values to add', values_to_add)
                for value in values_to_add:
                    index = pydash.find_index(current_schedules, lambda x: x['dayEN']==value['dayEN'] and x['startHour']==value['startHour'] and x['endHour']==value['endHour']);
                    if index == -1:
                        schedules.append(value);
            else:
                schedules = values_to_add.copy();

        
        # new_target = ArrayUtils.filter(current_targeted, lambda x: pydash.find(remove_on_targeted_locations, lambda y: x['criterionId']==y['criterionId']) is None) + targeted_locations;
        # new_excluded = ArrayUtils.filter(current_excluded, lambda x: pydash.find(remove_on_excluded_locations, lambda y: x['criterionId']==y['criterionId']) is None) + excluded_locations;
        if bool(targeted_locations) is True:
            data_add['targetedLocations'] = targeted_locations;
            data_final['targetedLocations'] = current_targeted + targeted_locations;
            print('targeted locations', targeted_locations)
        if bool(excluded_locations) is True:
            data_add['excludedLocations'] = excluded_locations;
            data_final['excludedLocations'] = current_excluded + excluded_locations;
        if bool(schedules) is True:
            data_add['adsSchedules'] = schedules;
            data_final['adsSchedules'] = current_schedules + schedules;
        print('data add', data_add)
        if bool(data_add) is False:
            return ApiResponse(ResponseStatus.ERROR, StatusCode.status_304, None, Error("No data to update","INVALID_REQUEST", 1));
        data_response = {"values": data_add, "combined": data_final};
        return ApiResponse(ResponseStatus.OK, StatusCode.status_304, data_response, None);
                        
                        
    def prepareRemoveCriterion(self, data):
        campaign_remove: CampaignDisplay = self.keep_only(CampaignDisplay.from_dict(data.copy()), pydash.keys(data.copy()));
        data_remove = {};
        data_final = {};
        current_targeted = self.locations_to_list(self.targetedLocations.copy());
        current_excluded = self.locations_to_list(self.excludedLocations.copy());
        current_schedules = self.schedules_to_list(self.adsSchedules.copy());
        targeted_locations = [];
        excluded_locations = [];
        schedules = [];
        if bool(campaign_remove.targetedLocations) is True:
            values_to_remove = self.locations_to_list(campaign_remove.targetedLocations);
            if bool(current_targeted) is True:
                for value in values_to_remove:
                    index = pydash.find_index(current_targeted, lambda x: x['criterionId']==value['criterionId']);
                    if index != -1:
                        targeted_locations.append(current_targeted.copy()[index]);
        if bool(campaign_remove.excludedLocations) is True:
            values_to_remove = self.locations_to_list(campaign_remove.excludedLocations);
            if bool(current_excluded) is True:
                for value in values_to_remove:
                    index = pydash.find_index(current_excluded, lambda x: x['criterionId']==value['criterionId']);
                    if index != -1:
                        excluded_locations.append(current_excluded.copy()[index]);
        
        if bool(campaign_remove.adsSchedules) is True:
            values_to_remove = self.locations_to_list(campaign_remove.excludedLocations);
            if bool(current_schedules) is True:
                for value in values_to_remove:
                    index = pydash.find_index(current_schedules, lambda x: x['dayEN']==value['dayEN'] and x['startHour']==value['startHour'] and x['endHour']==value['endHour']);
                    if index != -1:
                        schedules.append(current_schedules.copy()[index]);
        
        if bool(targeted_locations) is True:
            data_remove['targetedLocations'] = targeted_locations;
            data_final['targetedLocations'] = ArrayUtils.filter(current_targeted, lambda x: pydash.find(targeted_locations, lambda y: x['criterionId']==y['criterionId']) is None);
        if bool(excluded_locations) is True:
            data_remove['excludedLocations'] = excluded_locations;
            data_final['excludedLocations'] = ArrayUtils.filter(current_excluded, lambda x: pydash.find(excluded_locations, lambda y: x['criterionId']==y['criterionId']) is None);
        if bool(schedules) is True:
            data_remove['adsSchedules'] = schedules;
            data_final['adsSchedules'] = ArrayUtils.filter(current_schedules, lambda x: pydash.find(schedules, lambda y: x['dayEN']==y['dayEN'] and x['startHour']==y['startHour'] and x['endHour']==y['endHour']) is None);
        data_response = {"values": data_remove, "combined": data_final};
        if bool(data_remove) is False:
            return ApiResponse(ResponseStatus.ERROR, StatusCode.status_304, None, Error("No data to remove","INVALID_REQUEST", 1));
    
        return ApiResponse(ResponseStatus.OK, StatusCode.status_304, data_response, None);
        

    def getChangedFields(self, data):
        last_value = self.to_dict();
        filtered_value = pydash.pick(data, CampaignDisplayFields().filtered_keys('editable', True));
        new_value = DictUtils.merge_dict(filtered_value, last_value, True);
        _changed_fields = DictUtils.compare_nested_objects(last_value, new_value);
        changed_fields = ArrayUtils.filter(DictUtils.get_keys(_changed_fields), lambda x: str(x).find('.') == -1)
        return changed_fields
            
    def update(self, data, silent=False):
        # try:
        data_update = {};
        changed_fields = [];
        if silent is True:
            data_update = pydash.pick(data, CampaignDisplayFields().filtered_keys('pickable', True));
        else:
            last_value = self.to_dict();
            # print('last', last_value);
            filtered_value = pydash.pick(data, CampaignDisplayFields().filtered_keys('editable', True));
            # print('last', last_value['startDate']);
            # print('data', data['startDate']);
            # print('filtered', filtered_value['startDate']);
            new_value = DictUtils.merge_dict(filtered_value, last_value, True);
            _changed_fields = DictUtils.compare_nested_objects(last_value, new_value, '', ['criterionId']);
            # print('_changed_fields', DictUtils.get_keys(_changed_fields))
            changed_fields = ArrayUtils.filter(DictUtils.get_keys(_changed_fields), lambda x: str(x).find('.') == -1)
            data_update = DictUtils.dict_from_keys(filtered_value, changed_fields);
            # print('data update', data_update)
        if bool(data_update) is False:
            return ApiResponse(ResponseStatus.OK, StatusCode.status_200, data_update, None);
        # print('data_update', data_update)
        if 'startDate' in data_update or 'endDate' in data_update:

            dates = DateUtils.format_campaign_date(new_value, False)
            if dates.status == "error":
                return dates
        
            data_update = {**data_update, **dates.data}
            changed_fields.extend(DictUtils.get_keys(data_update))
        try:
            self.document_reference().set(data_update, merge=True)
        except Exception as e:
            return ApiResponse(ResponseStatus.ERROR, StatusCode.status_500, None, Error(str(e),"INTERNAL_ERROR", 0))
        # return DictUtils.dict_from_keys(self.get(), changed_fields);
        # print('fields', fields)
        # print('data_update', data_update)
        return ApiResponse(ResponseStatus.OK, StatusCode.status_200, DictUtils.pick_fields(data_update, CAMPAIGN_DISPLAY_PICKABLE_FIELDS), None);
        # return {"status": "ok", "data": CampaignDisplay.from_dict(campaign=data_update).to_dict()};
        # return {"status": "ok", "data": DictUtils.filter_by_fields(self.get().to_dict(), fields)};
        # except Exception as e:
        #     print(e)
        #     return None;

    def remove(self, only_mark_as_removed=True):
        try:
            if self.id is None:
                return ApiResponse(ResponseStatus.ERROR, StatusCode.status_400, None, Error(f"Cannot identify Campaign with id {self.id}","INVALID_REQUEST", 1));
            if only_mark_as_removed is True:
                self.document_reference().set({"is_removed": True}, merge=True)
            else:
                self.document_reference().delete();
            return ApiResponse(ResponseStatus.OK, StatusCode.status_200, {"message": f"Campaign {self.id} deleted"}, None);
        except:
            return ApiResponse(ResponseStatus.ERROR, StatusCode.status_400, None, Error(f"An error occurated while removing authorization code with id {self.id}","INVALID_REQUEST", 1));




from .ad import (ImageAd, ImageAdToPublish, filter_image_ad_request_fields)
from .ad_fields import (ImageAdFields, ImageAdFieldsProps, IMAGE_AD_PICKABLE_FIELDS, DISPLAY_ADS_COLLECTION)
from .native.ad import (ResponsiveDisplayAd, NativeAdsToPublish, filter_responsive_display_ad_request_fields)
from .native.ad_fields import (ResponsiveDisplayAdFields, ResponsiveDisplayAdFieldsProps, RESPONSIVE_DISPLAY_AD_PICKABLE_FIELDS, RESPONSIVE_DISPLAY_ADS_COLLECTION)



import uuid
from dataclasses import dataclass
from datetime import datetime, timezone
from typing import Any, Optional

from adafri.v1.base.firebase_collection import FirebaseCollectionBase

SCENARIO_COLLECTION = 'workflow_scenarios'

SCENARIO_PICKABLE_FIELDS = [
    'id', 'workspaceId', 'accountId', 'campaignId',
    'name', 'description', 'budget', 'bidStrategy',
    'targetAudience', 'expectedReach', 'expectedClicks', 'expectedConversions',
    'status', 'score', 'ownerId', 'allowedTeamIds', 'allowedMemberUids',
    'createdAt', 'updatedAt', 'createdBy',
]

SCENARIO_MUTABLE_FIELDS = [
    'name', 'description', 'budget', 'bidStrategy',
    'targetAudience', 'expectedReach', 'expectedClicks', 'expectedConversions',
    'status', 'score', 'allowedTeamIds', 'allowedMemberUids',
]

SCENARIO_STATUS = ['draft', 'selected', 'rejected']


def _now() -> str:
    return datetime.now(timezone.utc).isoformat()


def _new_id() -> str:
    return str(uuid.uuid4())


class WorkflowScenarioFields:
    @staticmethod
    def check_create_fields(data: dict) -> dict:
        if not data:
            return {'error': 'Missing data'}
        if not data.get('name', '').strip():
            return {'error': 'name is required'}
        if not data.get('workspaceId', '').strip():
            return {'error': 'workspaceId is required'}
        return data

    @staticmethod
    def mutable_keys() -> list:
        return SCENARIO_MUTABLE_FIELDS


@dataclass(init=False)
class WorkflowScenario(FirebaseCollectionBase):
    id: str
    workspaceId: str
    accountId: str
    campaignId: str
    name: str
    description: str
    budget: float
    bidStrategy: str
    targetAudience: dict
    expectedReach: int
    expectedClicks: int
    expectedConversions: int
    status: str
    score: float
    ownerId: str
    allowedTeamIds: list
    allowedMemberUids: list
    createdAt: str
    updatedAt: str
    createdBy: str

    def __init__(self, data=None, **kwargs):
        super().__init__(collection_name=SCENARIO_COLLECTION, **kwargs)
        d = data or {}
        self.id = d.get('id') or _new_id()
        self.workspaceId = d.get('workspaceId', '')
        self.accountId = d.get('accountId', '')
        self.campaignId = d.get('campaignId', '')
        self.name = d.get('name', '')
        self.description = d.get('description', '')
        self.budget = float(d.get('budget') or 0)
        self.bidStrategy = d.get('bidStrategy', 'TARGET_CPA')
        self.targetAudience = dict(d.get('targetAudience') or {})
        self.expectedReach = int(d.get('expectedReach') or 0)
        self.expectedClicks = int(d.get('expectedClicks') or 0)
        self.expectedConversions = int(d.get('expectedConversions') or 0)
        self.status = d.get('status', 'draft')
        self.score = float(d.get('score') or 0)
        self.ownerId = d.get('ownerId', '')
        self.allowedTeamIds = list(d.get('allowedTeamIds') or [])
        self.allowedMemberUids = list(d.get('allowedMemberUids') or [])
        self.createdAt = d.get('createdAt') or _now()
        self.updatedAt = d.get('updatedAt') or _now()
        self.createdBy = d.get('createdBy', '')

    @classmethod
    def from_dict(cls, obj: Any, **kwargs) -> 'WorkflowScenario':
        return cls(obj, **kwargs)

    def to_dict(self) -> dict:
        return {
            'id': self.id,
            'workspaceId': self.workspaceId,
            'accountId': self.accountId,
            'campaignId': self.campaignId,
            'name': self.name,
            'description': self.description,
            'budget': self.budget,
            'bidStrategy': self.bidStrategy,
            'targetAudience': self.targetAudience,
            'expectedReach': self.expectedReach,
            'expectedClicks': self.expectedClicks,
            'expectedConversions': self.expectedConversions,
            'status': self.status,
            'score': self.score,
            'ownerId': self.ownerId,
            'allowedTeamIds': self.allowedTeamIds,
            'allowedMemberUids': self.allowedMemberUids,
            'createdAt': self.createdAt,
            'updatedAt': self.updatedAt,
            'createdBy': self.createdBy,
        }

    @staticmethod
    def toListDict(items: list, fields=None) -> list:
        return [item.to_dict() for item in items]

    @classmethod
    def find(cls, scenario_id: str, account_id: str = None) -> 'Optional[WorkflowScenario]':
        inst = cls()
        doc = inst.collection().document(scenario_id).get()
        if not doc.exists:
            return None
        data = {**doc.to_dict(), 'id': doc.id}
        if account_id and data.get('accountId') != account_id:
            return None
        return cls(data)

    @classmethod
    def listByWorkspace(cls, workspace_id: str) -> 'list[WorkflowScenario]':
        inst = cls()
        docs = inst.collection().where('workspaceId', '==', workspace_id).stream()
        return [cls({**d.to_dict(), 'id': d.id}) for d in docs]

    def save(self) -> 'WorkflowScenario':
        self.collection().document(self.id).set(self.to_dict())
        return self

    def update(self, data: dict) -> 'WorkflowScenario':
        filtered = {k: v for k, v in data.items() if k in SCENARIO_MUTABLE_FIELDS}
        filtered['updatedAt'] = _now()
        self.collection().document(self.id).update(filtered)
        for k, v in filtered.items():
            setattr(self, k, v)
        return self

    def delete(self) -> None:
        self.collection().document(self.id).delete()



def cloud_firestore_query_formatter(key: str, value: str | bool | int | float | list, comp: str = "=="):
    return {key: {"value": value, "comp": comp}}

def cloud_firestore_query_builder(object: dict):
    query = []
    for key in object:
        data_value = object.get(key)
        comp = "=="
        value = None
        if isinstance(data_value, dict):
            value = data_value.get("value", None)
            comp = data_value.get("comp", "==")
        else:
            value = data_value
        if value is None:
            continue
        query.append({"key": key, "value": value, "comp": comp})
    return query



import random
import string
from dataclasses import dataclass
from datetime import datetime
from typing import Any

import pydash
from adafri.cache.lib import CacheClass
from adafri.v1 import version
from firebase_admin.firestore import firestore

from .....utils import (
    ArrayUtils,
    BaseClass,
    DictUtils,
    RequestFields,
    get_object_model_class,
    get_request_fields,
    init_class_kwargs,
)
from .....utils.response import ApiResponse, Error, ResponseStatus, StatusCode
from ...data.demographics import ages_data, genders_data
from ...data.devices import devices_data
from ...models.adgroups.base import BaseAdGroup, BaseAdGroupFieldsProps
from ...models.display_ads import DisplayAdsToPublish
from ...models.placement import Website, YoutubeChannel, YoutubeVideo
from ...models.user_interest import UserInterest
from ..ads.native import NativeAdsToPublish
from .adgroup_fields import (
    ADGROUP_DISPLAY_PICKABLE_FIELDS,
    DISPLAY_ADGROUPS_COLLECTION,
    AdGroupDisplayFields,
    AdGroupDisplayFieldsProps,
)


def filter_adgroup_display_request_fields(fields, default_fields = ADGROUP_DISPLAY_PICKABLE_FIELDS):
    request_fields = get_request_fields(fields, default_fields, [default_fields[0]])
    if request_fields is None or bool(request_fields) is False:
        request_fields = [default_fields[0]]
    return request_fields

@dataclass(init=False)
class AdGroupDisplay(BaseAdGroup):
    targetedPlacements: list[Website]
    excludedPlacements: list[Website]
    websites: list[Website]
    youtubeChannels: list[YoutubeChannel]
    youtubeVideos: list[YoutubeVideo]
    user_interest: list[UserInterest]
    __baseFields = AdGroupDisplayFieldsProps
    class_object = None;

    def __init__(self, adgroup=None, **kwargs):
        _adgroup = adgroup
        if type(adgroup) is str:
            _adgroup = {"id": adgroup}
        (cls_object, keys, data_args) = init_class_kwargs(self, _adgroup, ADGROUP_DISPLAY_PICKABLE_FIELDS, AdGroupDisplayFieldsProps, DISPLAY_ADGROUPS_COLLECTION, ['id'], **kwargs)
        super().__init__(**{**data_args, "adgroup": _adgroup});
        for key in keys:
            if BaseAdGroupFieldsProps.get(key) is None:
                website = ['targetedPlacements', 'excludedPlacements', 'websites']
                if key in website:
                    setattr(self, key, Website().fromListDict(cls_object[key]))
                elif key == 'youtubeChannels':
                    setattr(self, key, YoutubeChannel().fromListDict(cls_object[key]))
                elif key == 'youtubeVideos':    
                    setattr(self, key, YoutubeVideo().fromListDict(cls_object[key]))
                elif key == 'images':
                    setattr(self, key, DisplayAdsToPublish().fromListDict(cls_object[key]))
                elif key == 'imagesNative':
                    setattr(self, key, NativeAdsToPublish().fromListDict(cls_object[key]))
                elif key == 'user_interest':
                    setattr(self, key, UserInterest().fromListDict(cls_object[key]))
                else:
                    setattr(self, key, cls_object[key]) 
        self.class_object = AdGroupDisplay
        self.cache = CacheClass("adgroup_display", AdGroupDisplay)

    def placements_to_list(self, devices: list[Website] | None = None, fields = None):
        if devices is None:
            return []
        return Website().toListDict(devices, fields)
    def channels_to_list(self, channels: list[YoutubeChannel] | None = None, fields = None):
        if channels is None:
            return []
        return YoutubeChannel().toListDict(channels, fields)
    def videos_to_list(self, videos: list[YoutubeVideo] | None = None, fields = None):
        if videos is None:
            return []
        return YoutubeVideo().toListDict(videos, fields)
    def interest_to_list(self, interest: list[UserInterest] | None = None, fields = None):
        if interest is None:
            return []
        return UserInterest().toListDict(interest, fields)
    
    def list_to_placements(self, placements: list[Website] | None = None, fields = None):
        if placements is None:
            return []
        return Website().toListDict(placements, fields)
    def list_to_channels(self, channels: list[YoutubeChannel] | None = None, fields = None):
        if channels is None:
            return []
        return YoutubeChannel().toListDict(channels, fields)
    def list_to_videos(self, videos: list[YoutubeVideo] | None = None, fields = None):
        if videos is None:
            return []
        return YoutubeVideo().toListDict(videos, fields)
    def list_to_interests(self, interest: list[UserInterest] | None = None, fields = None):
        if interest is None:
            return []
        return UserInterest().toListDict(interest, fields)
    def is_published(self):
        return self.ad_group_id > 0
    def getChangedFields(self, data):
        last_value = self.to_dict();
        filtered_value = pydash.pick(data, AdGroupDisplayFields().filtered_keys('editable', True));
        new_value = DictUtils.merge_dict(filtered_value, last_value, True);
        _changed_fields = DictUtils.compare_nested_objects(last_value, new_value);
        changed_fields = ArrayUtils.filter(DictUtils.get_keys(_changed_fields), lambda x: str(x).find('.') == -1)
        return changed_fields
    def prepareAddCriterion(self, data):
        adgroup_add: AdGroupDisplay = self.keep_only(AdGroupDisplay.from_dict(data.copy()), pydash.keys(data.copy()));
        data_add = {};
        data_final = {};
        current_websites = self.placements_to_list(self.websites).copy();
        current_channels = self.channels_to_list(self.youtubeChannels).copy();
        current_videos = self.videos_to_list(self.youtubeVideos).copy();
        current_interests = self.interest_to_list(self.user_interest).copy();
        current_ages = self.ages_to_list(self.ages).copy()
        current_genders = self.genders_to_list(self.genders).copy()
        targeted_websites = [];
        targeted_channels = [];
        targeted_videos = []
        targeted_interests = []
        targeted_ages = []
        targeted_genders = []

        valid_campaign_status = ["enabled", "paused"]
        if bool(adgroup_add.name) is True and bool(adgroup_add.name.strip()) is True:
            data_add['name'] = adgroup_add.name;
            data_final['name'] = adgroup_add.name;
        if bool(adgroup_add.status) is True and adgroup_add.status.lower() in valid_campaign_status:
            data_add['status'] = adgroup_add.status;
            data_final['status'] = adgroup_add.status;

        if bool(adgroup_add.ages) is True:
            values_to_add = self.ages_to_list(adgroup_add.ages).copy();
            if bool(current_ages) is True:
                for value in values_to_add:
                    if pydash.find(current_ages, lambda x: x['criterionId']==value['criterionId']) is None:
                        targeted_ages.append(value);
            else:
                for value in values_to_add:
                    targeted_ages.append(value);
        
        if bool(adgroup_add.genders) is True:
            values_to_add = self.genders_to_list(adgroup_add.genders_to_list).copy();
            if bool(current_genders) is True:
                for value in values_to_add:
                    if pydash.find(current_genders, lambda x: x['criterionId']==value['criterionId']) is None:
                        targeted_genders.append(value);
            else:
                for value in values_to_add:
                    targeted_genders.append(value);
        
        if bool(adgroup_add.websites) is True:
            values_to_add = self.placements_to_list(adgroup_add.websites);
            if bool(current_websites) is True:
                for value in values_to_add:
                    if pydash.find(current_websites, lambda x: x['url']==value['url']) is None:
                        targeted_websites.append(value);
            else:
                for value in values_to_add:
                    targeted_websites.append(value);
        
        if bool(adgroup_add.youtubeChannels) is True:
            values_to_add = self.channels_to_list(adgroup_add.youtubeChannels);
            if bool(current_channels) is True:
                for value in values_to_add:
                    if pydash.find(current_channels, lambda x: x['id']==value['id']) is None:
                        targeted_channels.append(value);
            else:
                for value in values_to_add:
                    targeted_channels.append(value);
        
        if bool(adgroup_add.youtubeVideos) is True:
            values_to_add = self.videos_to_list(adgroup_add.youtubeVideos);
            if bool(current_videos) is True:
                for value in values_to_add:
                    if pydash.find(current_videos, lambda x: x['id']==value['id']) is None:
                        targeted_videos.append(value);
            else:
                for value in values_to_add:
                    targeted_videos.append(value);
        
        if bool(adgroup_add.user_interest) is True:
            values_to_add = self.interest_to_list(adgroup_add.user_interest);
            if bool(current_interests) is True:
                for value in values_to_add:
                    if pydash.find(current_interests, lambda x: x['id']==value['id']) is None:
                        targeted_interests.append(value);
            else:
                for value in values_to_add:
                    targeted_interests.append(value);

        if bool(targeted_websites) is True:
            data_add['websites'] = targeted_websites.copy();
            data_final['websites'] = current_websites.copy() + targeted_websites.copy();
            # print('targeted websites', websites)
        if bool(targeted_channels) is True:
            data_add['youtubeChannels'] = targeted_channels.copy();
            data_final['youtubeChannels'] = current_channels.copy() + targeted_channels.copy();
        
        if bool(targeted_videos) is True:
            data_add['youtubeVideos'] = targeted_videos.copy();
            data_final['youtubeVideos'] = current_videos.copy() + targeted_videos.copy();
        
        if bool(targeted_interests) is True:
            data_add['user_interest'] = targeted_interests.copy();
            data_final['user_interest'] = current_interests.copy() + targeted_interests.copy();

        print('data add', data_add)
        if bool(data_add) is False:
            return ApiResponse(ResponseStatus.ERROR, StatusCode.status_304, None, Error("No data to update","INVALID_REQUEST", 1));
        data_response = {"values": data_add, "combined": data_final};
        return ApiResponse(ResponseStatus.OK, StatusCode.status_304, data_response, None)
    
    def prepareRemoveCriterion(self, data):
        adgroup_remove: AdGroupDisplay = self.keep_only(AdGroupDisplay.from_dict(data.copy()), pydash.keys(data.copy()));
        data_remove = {};
        data_final = {};
        current_websites = self.placements_to_list(self.websites).copy();
        current_channels = self.channels_to_list(self.youtubeChannels).copy();
        current_videos = self.videos_to_list(self.youtubeVideos).copy();
        current_interests = self.interest_to_list(self.user_interest).copy();
        current_ages = self.ages_to_list(self.ages).copy()
        current_genders = self.genders_to_list(self.genders).copy()
        targeted_websites = [];
        targeted_channels = [];
        targeted_videos = []
        targeted_interests = []
        targeted_ages = []
        targeted_genders = []
        if bool(adgroup_remove.websites) is True:
            values_to_remove = self.placements_to_list(adgroup_remove.websites).copy();
            if bool(current_websites) is True:
                for value in values_to_remove:
                    index = pydash.find_index(current_websites, lambda x: x['id']==value['id']);
                    if index != -1:
                        targeted_websites.append(current_websites.copy()[index]);
        if bool(adgroup_remove.youtubeChannels) is True:
            values_to_remove = self.channels_to_list(adgroup_remove.youtubeChannels).copy();
            if bool(current_channels) is True:
                for value in values_to_remove:
                    index = pydash.find_index(current_channels, lambda x: x['id']==value['id']);
                    if index != -1:
                        targeted_channels.append(current_channels.copy()[index]);
        if bool(adgroup_remove.youtubeVideos) is True:
            values_to_remove = self.videos_to_list(adgroup_remove.youtubeVideos).copy();
            if bool(current_videos) is True:
                for value in values_to_remove:
                    index = pydash.find_index(current_videos, lambda x: x['id']==value['id']);
                    if index != -1:
                        targeted_videos.append(current_videos.copy()[index]);
        
        if bool(adgroup_remove.user_interest) is True:
            values_to_remove = self.videos_to_list(adgroup_remove.user_interest).copy();
            if bool(current_interests) is True:
                for value in values_to_remove:
                    index = pydash.find_index(current_interests, lambda x: x['id']==value['id']);
                    if index != -1:
                        targeted_interests.append(current_interests.copy()[index]);
        
        if bool(targeted_websites) is True:
            data_remove['websites'] = targeted_websites.copy();
            data_final['websites'] = ArrayUtils.filter(current_websites, lambda x: pydash.find(targeted_websites, lambda y: x['url']==y['url']) is None);
        if bool(targeted_channels) is True:
            data_remove['youtubeChannels'] = targeted_channels.copy();
            data_final['youtubeChannels'] = ArrayUtils.filter(current_channels, lambda x: pydash.find(targeted_channels, lambda y: x['id']==y['id']) is None);
        if bool(targeted_videos) is True:
            data_remove['youtubeVideos'] = targeted_videos.copy();
            data_final['youtubeVideos'] = ArrayUtils.filter(current_videos, lambda x: pydash.find(targeted_videos, lambda y: x['id']==y['id']) is None);
        if bool(targeted_interests) is True:
            data_remove['user_interest'] = targeted_interests.copy();
            data_final['user_interest'] = ArrayUtils.filter(current_interests, lambda x: pydash.find(targeted_interests, lambda y: x['id']==y['id']) is None);

        data_response = {"values": data_remove, "combined": data_final};
        if bool(data_remove) is False:
            return ApiResponse(ResponseStatus.ERROR, StatusCode.status_304, None, Error("No data to remove","INVALID_REQUEST", 1));
    
        return ApiResponse(ResponseStatus.OK, StatusCode.status_304, data_response, None);

    @staticmethod
    def generate_model(_key_="default_value"):
        model = {};
        props = AdGroupDisplayFieldsProps
        for k in DictUtils.get_keys(props):
            if k=='ages':
                model[k] = ages_data;
                continue
            if k=='genders':
                model[k] = genders_data;
                continue
            if k=='devicesTargeted':
                model[k] = devices_data;
                continue
            if k=='api_version':
                model[k] = version;
                continue;
            if _key_ in props[k] and props[k][_key_] is not None:
                model[k] = props[k][_key_];
        
        # Generate random campaign name
        random_suffix = ''.join(random.choices(string.ascii_uppercase + string.digits, k=6))
        model['name'] = f'Campaign Display {random_suffix}'
        model['startDate'] = datetime.now().strftime('%Y-%m-%d')
        return model;

    @staticmethod
    def from_dict(adgroup: Any, db=None, collection_name=None) -> 'AdGroupDisplay':
        cls_object, keys = get_object_model_class(adgroup, AdGroupDisplay, AdGroupDisplayFieldsProps);
        _adgroup = AdGroupDisplay(cls_object, db=db, collection_name=collection_name)
        return _adgroup
    def generateAdgroup(self, campaign, data=None):
        if campaign is None:
            return ApiResponse(ResponseStatus.ERROR, StatusCode.status_400, None, Error("Campaign not found","INVALID_REQUEST", 1))

        model = AdGroupDisplay.generate_model()
        model['accountId'] = campaign.accountId
        model['clientCustomerId'] = campaign.clientCustomerId
        if data is not None:
            model = DictUtils.merge_dict(data, model, True)
        # print('create adgroup with model', model)
        create = AdGroupDisplay().create(model, campaign)
        if create.status == 'error':
            return create
        # print('fields', CAMPAIGN_DISPLAY_PICKABLE_FIELDS)
        values = create.data
        # values = DictUtils.omit_fields(values, ADGROUP_DISPLAY_PICKABLE_FIELDS + ['owner', 'id_campagne', 'ad_group_id'])
        # print('values', values)
        return ApiResponse(ResponseStatus.OK, StatusCode.status_200, values, None)
    
    @staticmethod
    def property_keys():
        keys = [AdGroupDisplayFields.ages, 
                AdGroupDisplayFields.genders, 
                AdGroupDisplayFields.devicesTargeted, 
                AdGroupDisplayFields.devicesExcluded, 
                AdGroupDisplayFields.websites,
                AdGroupDisplayFields.youtubeChannels,
                AdGroupDisplayFields.youtubeVideos,
                AdGroupDisplayFields.images,
                AdGroupDisplayFields.imagesNative,
                AdGroupDisplayFields.user_interest]
        return keys
    @staticmethod
    def get_ad_group_props(value):
        if value is None:
            return {}
        return DictUtils.pick_fields(value, AdGroupDisplay.property_keys())
    def to_dict(self, fields=None):
        base = super().to_dict(fields)
        return DictUtils.remove_none_values({
            **base,
            "targetedPlacements": Website().toListDict(self.targetedPlacements, None),
            "excludedPlacements": Website().toListDict(self.excludedPlacements, None),
            "websites": Website().toListDict(self.websites, None),
            "youtubeChannels": YoutubeChannel().toListDict(self.youtubeChannels, None),
            "youtubeVideos": YoutubeVideo().toListDict(self.youtubeVideos, None),
            "images": DisplayAdsToPublish().toListDict(getattr(self, 'images', None), None),
            "imagesNative": NativeAdsToPublish().toListDict(getattr(self, 'imagesNative', None), None),
            "user_interest": UserInterest().toListDict(getattr(self, 'user_interest', None), None),
    })
    @staticmethod
    def findAdgroup(id, aacid=None):
        isID = False
        try:
            id = int(id)
        except Exception as e:
            isID = True
        if isID is True:
            campaign = AdGroupDisplay.from_dict({"id": id}).get(ADGROUP_DISPLAY_PICKABLE_FIELDS)
            if campaign is None:
                return None
            return campaign 
        if id is None or aacid is None or len(id) == 0 or len(aacid) == 0:
            return None
        campaign = AdGroupDisplay(None).query(AdGroupDisplay, "adgroup", query_params=[{"key": "accountId", "comp": "==", "value": aacid}, {"key": "ad_group_id", "comp": "==", "value": id}], first=True)
        print('finding', id)
        return campaign
    def getAdGroups(self, accountId, campaignId, fields=['id', 'name'], response_type = 'dict'):
        isID = False
        try:
            campaignId = int(campaignId)
        except:
            isID = True
        if isID is False and accountId is None:
            return None
        adgroups = []
        queryParams = [{"key": "accountId", "comp": "==", "value": accountId}, ]
        if isID is False:
            queryParams.append({"key": "campaign_id", "comp": "==", "value": campaignId})
        else :
            queryParams.append({"key": "campaign_id_parent", "comp": "==", "value": campaignId})
        adgroups_ = self.query(AdGroupDisplay, "adgroup", query_params=queryParams)
        
        request_fields = filter_adgroup_display_request_fields(fields)
        for adgroup in adgroups_:
            if response_type == 'class':
                adgroups.append(adgroup)
            else:
                if fields is None:
                    adgroups.append(adgroup.to_dict())
                    continue
                adgroups.append(DictUtils.filter_by_fields(adgroup.to_dict(), request_fields))
        return adgroups
    
    def get(self, fields=None):
        if bool(self.id) is None:
            return None;
        adgroup_cache: AdGroupDisplay | None = self.cache.get(self.id);
        if adgroup_cache is not None:
            return adgroup_cache
        request_fields = filter_adgroup_display_request_fields(fields)
        doc = self.document_reference().get();
        if doc.exists is False:
            return None;
        doc_dict = doc.to_dict()
        if 'id' in doc_dict and bool(doc_dict['id']) is False:
            del doc_dict['id']
        data = {"id": doc.id, **doc_dict}
        adgroup = AdGroupDisplay.from_dict(adgroup=AdGroupDisplay.from_dict(adgroup=data).to_dict(),  db=self.db, collection_name=self.collection_name)
        self.cache.set(adgroup.id, adgroup);
        return adgroup;

    def getByUserId(self,id, fields=None):
        adgroups = [];
        request_fields = filter_adgroup_display_request_fields(fields)
        docs = self.collection().where('owner','==',id).get();
        for doc in docs:
            data = doc.to_dict();
            data['id'] = doc.id;
            adgroups.append(AdGroupDisplay.from_dict(adgroup=AdGroupDisplay.from_dict(adgroup=data).to_json(request_fields),  db=self.db, collection_name=self.collection_name));
        return adgroups;
    def normalize_response(self):
        data = AdGroupDisplay.from_dict(DictUtils.pick_fields(self.to_dict(), ADGROUP_DISPLAY_PICKABLE_FIELDS))
        return AdGroupDisplay.from_dict(DictUtils.omit_fields(data.to_dict(), ["owner", "createdBy"]))

    def create(self, data, campaign, only_generate = False):
        # creation_date = created_at * 1000
        # return DateUtils.from_timestamp(creation_date).isoformat()
        # try:
        # print('campaign account', data)
        data_create = DictUtils.dict_from_keys(data, AdGroupDisplayFields().filtered_keys('mutable', True));
        if bool(data_create) is False:
            return ApiResponse(ResponseStatus.ERROR, StatusCode.status_400, None, Error("Invalid request","INVALID_REQUEST", 1))
        data_create = DictUtils.merge_dict({
                "provider": campaign.provider,
                "accountId": campaign.accountId,
                "owner": campaign.owner,
                "clientCustomerId": campaign.clientCustomerId,
                "campaign_id_parent": campaign.id,
                "createdBy": f"users/{campaign.owner}"
            }, data_create,True)
        # print('data_create', data_create)
        document_id = self.collection().document().id
        query_filter = [{"key": "name", "comp": "==", "value": data_create['name']}, {"key": "accountId", "comp": "==", "value": campaign.accountId}, {"key": "clientCustomerId", "comp": "==", "value": campaign.clientCustomerId}];
        exist = self.query(AdGroupDisplay, "adgroup", query_filter, True)
        if exist is not None:
            return ApiResponse(ResponseStatus.ERROR, StatusCode.status_400, exist, Error(f"Adgroup name {data_create['name']} already exists","DUPLICATE_ENTITY", 2))
        model = AdGroupDisplay.generate_model()
        model['id'] = document_id
        data_create = DictUtils.merge_dict(data_create, model, True)
        create = {**data_create, 'createdAt': firestore.SERVER_TIMESTAMP}
        if only_generate is False or only_generate is None:
            doc_ref = self.collection().document(document_id);
            doc_ref.set(create, True);
        adgroup = AdGroupDisplay.from_dict(create).normalize_response()
        if adgroup is None:
            return ApiResponse(ResponseStatus.ERROR, StatusCode.status_400, None, Error("Failed to create adgroup","INVALID_REQUEST", 1))
        return ApiResponse(ResponseStatus.OK, StatusCode.status_200, adgroup.normalize_response(), None);  # Return the created campaign object.  # Return the created campaign object.  # Return the created campaign object.  # Return the created campaign object.  # Return the created campaign object.  # Return the created campaign object.  # Return the created campaign object.  # Return the created campaign object.  # Return the created campaign object.  # Return the created campaign object.  # Return the created campaign object.  # Return the
        # except Exception as e:
        #     print(e)
        #     return {"error": "Exception creating campaign"};

    def update(self, data, silent = True):
        try:
            data_update = {};
            changed_fields = [];
            if silent is True:
                data_update = pydash.pick(data, AdGroupDisplayFields().filtered_keys('pickable', True));
            else:
                last_value = self.to_dict();
                print('data update', data)
                # print('last', last_value);
                filtered_value = pydash.pick(data, AdGroupDisplayFields().filtered_keys('editable', True));
                # print('filtered', filtered_value);
                new_value = DictUtils.merge_dict(filtered_value, last_value, True);
                _changed_fields = DictUtils.compare_nested_objects(last_value, new_value);
                changed_fields = ArrayUtils.filter(DictUtils.get_keys(_changed_fields), lambda x: str(x).find('.') == -1)
                data_update = DictUtils.dict_from_keys(filtered_value, changed_fields);
            if bool(data_update) is False:
                return ApiResponse(ResponseStatus.OK, StatusCode.status_200, data_update, None);
            # try:
            #     self.document_reference().set(data_update, merge=True)
            # except Exception as e:
            #     return ApiResponse(ResponseStatus.ERROR, StatusCode.status_400, None, Error(str(e),"INVALID_REQUEST", 1))
            # return DictUtils.dict_from_keys(self.get(), changed_fields);
            # fields = changed_fields
            print('data_update adgroup', data_update)
            return ApiResponse(ResponseStatus.OK, StatusCode.status_200, DictUtils.pick_fields(data_update, ADGROUP_DISPLAY_PICKABLE_FIELDS), None);
        except Exception as e:
            print(e)
            return None;

    def to_json(self, _fields=None):
        return self.to_dict()
    def remove(self, only_mark_as_removed=True):
        if self.is_removed is True:
            return ApiResponse(ResponseStatus.ERROR, StatusCode.status_500, None, Error(f"Adgroup {self.id} already removed","INVALID_REQUEST", 1));
        try:
            if self.id is None:
                return ApiResponse(ResponseStatus.ERROR, StatusCode.status_400, None, Error(f"Cannot identify Campaign with id {self.id}","INVALID_REQUEST", 1));
            if only_mark_as_removed:
                self.document_reference().set({"is_removed": True}, merge=True)
            else:
                self.document_reference().delete();
            return ApiResponse(ResponseStatus.OK, StatusCode.status_200, {"message": f"Campaign {self.id} deleted"}, None);
        except:
            return ApiResponse(ResponseStatus.ERROR, StatusCode.status_400, None, Error(f"An error occurated while removing authorization code with id {self.id}","INVALID_REQUEST", 1));




import hashlib
import secrets
from dataclasses import dataclass
from datetime import datetime, timedelta, timezone
from typing import Any

from .....utils import DictUtils, Crypto, get_object_model_class, init_class_kwargs
from ....base.firebase_collection import FirebaseCollectionBase, getTimestamp
from .device_code_fields import DeviceCodeFields, DeviceCodeFieldsProps, STANDARD_FIELDS, DEVICE_CODE_COLLECTION
from .....utils.response import ApiResponse, Error, ResponseStatus, StatusCode


def _hash(value: str) -> str:
    return hashlib.sha256(value.encode()).hexdigest()


def _generate_user_code() -> str:
    part1 = secrets.token_hex(2).upper()
    part2 = secrets.token_hex(2).upper()
    return f"{part1}-{part2}"


@dataclass(init=False)
class DeviceCode(FirebaseCollectionBase):
    id: str
    device_code: str
    user_code: str
    client_id: str
    scope: str
    user_uid: str
    status: str
    interval: int
    expires_at: str
    last_polled_at: str
    created_at: str

    def __init__(self, dc=None, **kwargs):
        if type(dc) is str:
            dc = {"id": dc}
        (cls_object, keys, data_args) = init_class_kwargs(self, dc, STANDARD_FIELDS, DeviceCodeFieldsProps, DEVICE_CODE_COLLECTION, ['id'], **kwargs)
        super().__init__(**data_args)
        for key in keys:
            setattr(self, key, cls_object[key])

    @staticmethod
    def generate_model(_key_="default_value"):
        model = {}
        props = DeviceCodeFieldsProps
        for k in DictUtils.get_keys(props):
            model[k] = props[k][_key_]
        return model

    @staticmethod
    def from_dict(dc: Any = None, db=None, collection_name=None) -> 'DeviceCode':
        cls_object, keys = get_object_model_class(dc, DeviceCode, DeviceCodeFieldsProps)
        return DeviceCode(cls_object, db=db, collection_name=collection_name)

    def query(self, query_params: list, first=False, limit=None):
        try:
            result = []
            query_result = self.custom_query(query_params, first=first, limit=limit)
            if bool(query_result):
                if first:
                    return DeviceCode.from_dict(dc=query_result, db=self.db, collection_name=self.collection_name)
                else:
                    for doc in query_result:
                        result.append(DeviceCode.from_dict(dc=doc, db=self.db, collection_name=self.collection_name))
                    return result
            return None if first else []
        except Exception as e:
            print('Exception while querying DeviceCode', e)
            return None if first else []

    def to_dict(self):
        return {
            "id": self.id,
            "device_code": self.device_code,
            "user_code": self.user_code,
            "client_id": self.client_id,
            "scope": self.scope,
            "user_uid": self.user_uid,
            "status": self.status,
            "interval": self.interval,
            "expires_at": self.expires_at,
            "last_polled_at": self.last_polled_at,
            "created_at": self.created_at,
        }

    def to_json(self, _fields=None):
        return self.to_dict()

    def get_by_user_code(self, code: str) -> 'DeviceCode':
        return self.query([{"key": DeviceCodeFields.user_code, "comp": "==", "value": code}], first=True)

    def get_by_device_code(self, raw_device_code: str) -> 'DeviceCode':
        hashed = _hash(raw_device_code)
        return self.query([{"key": DeviceCodeFields.device_code, "comp": "==", "value": hashed}], first=True)

    def is_expired(self) -> bool:
        try:
            expires = datetime.fromisoformat(self.expires_at)
            return expires < datetime.now(timezone.utc)
        except Exception:
            return True

    def authorize(self, user_uid: str):
        self.document_reference(self.id).set({
            "user_uid": user_uid,
            "status": "authorized"
        }, merge=True)

    def poll(self) -> dict:
        now = datetime.now(timezone.utc)
        if self.last_polled_at:
            try:
                last = datetime.fromisoformat(self.last_polled_at)
                elapsed = (now - last).total_seconds()
                if elapsed < self.interval:
                    return {"error": "slow_down", "interval": self.interval}
            except Exception:
                pass
        self.document_reference(self.id).set({"last_polled_at": now.isoformat()}, merge=True)
        if self.is_expired():
            return {"error": "expired_token"}
        if self.status == "pending":
            return {"error": "authorization_pending"}
        if self.status == "denied":
            return {"error": "access_denied"}
        return {"status": self.status, "user_uid": self.user_uid}

    @classmethod
    def create(cls, client_id: str, scope: str = "", expires_in: int = 1800) -> tuple:
        raw_device_code = secrets.token_hex(32)
        hashed_device_code = _hash(raw_device_code)
        user_code = _generate_user_code()
        now = datetime.now(timezone.utc)
        expires_at = (now + timedelta(seconds=expires_in)).isoformat()
        model = {
            "device_code": hashed_device_code,
            "user_code": user_code,
            "client_id": client_id,
            "scope": scope,
            "user_uid": "",
            "status": "pending",
            "interval": 5,
            "expires_at": expires_at,
            "last_polled_at": "",
            "created_at": now.isoformat(),
        }
        obj = cls.from_dict(model)
        obj.id = Crypto().generate_id(client_id + "~device~" + user_code)
        doc_ref = DeviceCode(obj.to_json()).document_reference()
        doc_ref.set({**obj.to_json(), "createdAt": getTimestamp()}, merge=True)
        return obj, raw_device_code
